# -*- coding: utf-8 -*-

# ******************************************************************************
#                          S-PLUS CALIBRATION PIPELINE
#                               fit_zp_offsets.py
#             Compares two calibrations to derive zero-point offsets
# ******************************************************************************

"""
--------------------------------------------------------------------------------
   FUNCTIONS:
--------------------------------------------------------------------------------
create_compare_zp_file()


--------------------------------------------------------------------------------
   COMMENTS:
--------------------------------------------------------------------------------
This script should be run individually and not within the pipeline.py script

--------------------------------------------------------------------------------
   USAGE:
--------------------------------------------------------------------------------
$python3 fit_zp_offsets.py *field_list_file*
                           *calib1_suffix* *calib2_suffix*
                           *calib1_step* *calib2_step*
                           *calib1_path* *calib2_path*[opt]

----------------
"""

################################################################################
# Import external packages

import os
import sys

addsteps_path = os.path.split(__file__)[0]
pipeline_path = os.path.split(addsteps_path)[0]
spluscalib_path = os.path.split(pipeline_path)[0]

sys.path.append(spluscalib_path)

################################################################################
# Import spluscalib packages

from spluscalib import utils as ut

################################################################################
# Read parameters

field_list_file = sys.argv[1]

suffix1 = sys.argv[2]
suffix2 = sys.argv[3]

calib1_step = sys.argv[4]
calib2_step = sys.argv[5]

calib1_path = sys.argv[6]

try:
    calib2_path = sys.argv[7]
except IndexError:
    calib2_path = calib1_path


save_path = os.path.join(calib1_path, f'zp_offset_{suffix1}-{suffix2}')
ut.makedir(save_path)

################################################################################
# Initiate log file

log_file_name = os.path.join(save_path, 'fit_zp_offsets.log')
log_file_name = ut.gen_logfile_name(log_file_name)
log_file = os.path.join(save_path, log_file_name)

with open(log_file, "w") as log:
    log.write("")

################################################################################
# Reading field list

ut.printlog("*********** Reading field list **********", log_file)

fields = ut.load_field_list(field_list_file)

ut.printlog(f"Running the script for fields:", log_file)
ut.printlog(f"{fields}", log_file)

################################################################################
# Begin script

# ***************************************************
#    Create comparison file
# ***************************************************


def create_compare_zp_file():

    """
    Creates a file with zero points from calibrations 1 and 2
    """

    print("")
    ut.printlog(('********** '
                 'Creating zero-points comparation file '
                 '**********'),
                log_file)
    print("")

    save_name = f"zps_{calib1_step}1-{calib2_step}2.csv"
    save_file = os.path.join(save_path, save_name)

    if not os.path.exists(save_file):

        fields_zps = {}

        for field in fields:

            ####################################
            # Get calibration 1 zp file name base

            if calib1_step == 'external':
                zp_file1 = os.path.join(calib1_path, field,
                                        f"Calibration_{suffix1}",
                                        'external', f'{field}_external.zp')

            elif calib1_step == 'stlocus':
                zp_file1 = os.path.join(calib1_path, field,
                                  f"Calibration_{suffix1}",
                                  'stlocus', f'{field}_external_and_stlocus.zp')

            elif calib1_step == 'internal':
                zp_file1 = os.path.join(calib1_path, field,
                                  f"Calibration_{suffix1}",
                                'internal', f'{field}_external_and_internal.zp')

            elif calib1_step == 'final':
                zp_file1 = os.path.join(calib1_path, field,
                                  f"Calibration_{suffix1}",
                                       f'{field}_final.zp')

            else:
                zp_file1 = None

            ####################################
            # Get calibration 1 zp file name base

            if calib2_step == 'external':
                zp_file2 = os.path.join(calib2_path, field,
                                  f"Calibration_{suffix2}",
                                  'external', f'{field}_external.zp')

            elif calib2_step == 'stlocus':
                zp_file2 = os.path.join(calib2_path, field,
                                  f"Calibration_{suffix2}",
                                  'stlocus', f'{field}_external_and_stlocus.zp')

            elif calib2_step == 'internal':
                zp_file2 = os.path.join(calib2_path, field,
                                  f"Calibration_{suffix2}",
                                'internal', f'{field}_external_and_internal.zp')

            elif calib2_step == 'final':
                zp_file2 = os.path.join(calib2_path, field,
                                  f"Calibration_{suffix2}",
                                       f'{field}_final.zp')

            else:
                zp_file2 = None

            ############################
            # Fill dictionary fields_zps

            fields_zps[field] = [zp_file1, zp_file2]

        ########################
        # Get comparison catalog

        ut.zp_comparison(fields_zps  = fields_zps,
                         save_file   = save_file,
                         fields_list = fields)

        ut.printlog(f"Created file {save_file}", log_file)

    else:
        ut.printlog(f"Catalog {save_name} already exists", log_file)


create_compare_zp_file()


# ***************************************************
#    Fit zp offsets
# ***************************************************

def fit_zp_offsets():

    """
    Fits the offsets between the zero-points of two different calibrations
    """

    print("")
    ut.printlog(('********** '
                 'Fitting zero-points offsets '
                 '**********'),
                log_file)
    print("")

    zpcomp_name = f"zps_{calib1_step}1-{calib2_step}2.csv"
    zpcomp_file = os.path.join(save_path, zpcomp_name)

    save_name = f"offsets.zp"
    save_file = os.path.join(save_path, save_name)

    filters = ['SPLUS_U', 'SPLUS_F378', 'SPLUS_F395', 'SPLUS_F410',
               'SPLUS_F430', 'SPLUS_G', 'SPLUS_F515', 'SPLUS_R',
               'SPLUS_F660', 'SPLUS_I', 'SPLUS_F861', 'SPLUS_Z']

    if not os.path.exists(save_file):
        ut.zp_fit_offsets(zp_comparison_file = zpcomp_file,
                                   save_file = save_file,
                                     filters = filters)

        ut.printlog(f"Created file {save_file}", log_file)

    else:
        ut.printlog(f"ZP file {save_name} already exists", log_file)


fit_zp_offsets()



# -*- coding: utf-8 -*-

# ******************************************************************************
#                          S-PLUS CALIBRATION PIPELINE
#                         compute_XY_correction_maps.py
#      Computes XY zero-points offsets in relation to the reference catalog
# ******************************************************************************

"""
Combines intermediare calibration catalogs for a list of fields

Command line arguments for this script are, respectivelly:
1) Location of the S-PLUS fields list file
2) Location of the configuration file used in the calibration
3) Location to save the intermediary calibration catalogs

--------------------------------------------------------------------------------
   FUNCTIONS:
--------------------------------------------------------------------------------

concatenate_instrumental_catalogs()
concatenate_external_calib_catalogs()
concatenate_stlocus_calib_catalogs()
concatenate_internal_calib_catalogs()

--------------------------------------------------------------------------------
   COMMENTS:
--------------------------------------------------------------------------------
This script should be run individually and not within the pipeline.py script

--------------------------------------------------------------------------------
   USAGE:
--------------------------------------------------------------------------------
$python3 combine_calibration_catalogs.py *field_list_file* *config_file*

----------------
"""

################################################################################
# Import external packages

import os
import sys

addsteps_path = os.path.split(__file__)[0]
pipeline_path = os.path.split(addsteps_path)[0]
spluscalib_path = os.path.split(pipeline_path)[0]

sys.path.append(spluscalib_path)

################################################################################
# Import spluscalib packages

from spluscalib import utils as ut

################################################################################
# Read parameters

field_list_file = sys.argv[1]
conf_file = sys.argv[2]
save_path = sys.argv[3]

conf = ut.pipeline_conf(conf_file)

suffix = ut.calibration_suffix(conf)

if conf['reference_catalog'][0].lower() != 'ivezic':
    raise ValueError('Only Ivezic catalog is supported for XY computation')

################################################################################
# Initiate log file

ut.makedir(save_path)

log_file_name = os.path.join(save_path, 'compute_XY_maps.log')
log_file_name = ut.gen_logfile_name(log_file_name)
log_file = os.path.join(save_path, log_file_name)

with open(log_file, "w") as log:
    log.write("")

################################################################################
# Reading field list

ut.printlog("*********** Reading field list **********", log_file)

fields = ut.load_field_list(field_list_file)

ut.printlog(f"Running the pipeline for fields:", log_file)
ut.printlog(f"{fields}", log_file)

mag_pairs = {'SPLUS_U': 'SDSS_U',
             'SPLUS_G': 'SDSS_G',
             'SPLUS_R': 'SDSS_R',
             'SPLUS_I': 'SDSS_I',
             'SPLUS_Z': 'SDSS_Z'}

xbins = [0, 9200, 32]
ybins = [0, 9200, 32]

################################################################################
# Begin script

plots_path = os.path.join(save_path, 'plots')
ut.makedir(plots_path)

# ***************************************************
#    Convert catalogs to fits
# ***************************************************

def convert_2_fits():

    """
    Converts catalogs from ascii to fits
    """

    for field in fields:
        cat_name = f'{field}_mag_ext.cat'
        cat_step = 'external'

        cat_file = os.path.join(conf['save_path'], f'{field}',
                                f'Calibration_{suffix}', cat_step,
                                cat_name)

        save_name = f'{field}_mag_ext.fits'
        save_file = os.path.join(save_path, save_name)

        if not os.path.exists(save_file):

            cmd = f"java -jar {conf['path_to_stilts']} tcopy "
            cmd += f"in={cat_file} ifmt=ascii "
            cmd += f"out={save_file} ofmt=fits"

            ut.printlog(cmd, log_file)
            os.system(cmd)

        else:
            ut.printlog(f"File {save_file} already exists.", log_file)


convert_2_fits()

# ***************************************************
#    fix X,Y rotation
# ***************************************************

def fix_XY_rotation():

    """
    Fixes XY rotation of S-PLUS images
    """

    for field in fields:

        cat_name = f'{field}_mag_ext.fits'
        cat_file = os.path.join(save_path, cat_name)

        save_name = f'{field}_mag_ext_xycorr.fits'
        save_file = os.path.join(save_path, save_name)

        if not os.path.exists(save_file):

            ut.fix_xy_rotation(catalog = cat_file,
                               save_file = save_file,
                               xcol = 'X',
                               ycol = 'Y')

            ut.printlog(f"Created file {save_file}", log_file)

        else:
            ut.printlog(f"File {save_file} already exists.", log_file)


fix_XY_rotation()

# ***************************************************
#    Combine calibrated catalogs
# ***************************************************

def concatenate_calib_catalogs():

    """
    Concatenates calibration catalogs after the external calibration step
    """

    print("")
    ut.printlog(('********** '
                 'Concatenating external calibrated magnitudes '
                 '**********'),
                 log_file)
    print("")

    cat_name = '{field}_mag_ext_xycorr.fits'
    cat_file = os.path.join(save_path, cat_name)

    # Start save name
    save_name = 'concat_mag_ext_xycorr.cat'
    save_file = os.path.join(save_path, save_name)

    if not os.path.exists(save_file):

        # Create list of field catalogs
        catalogs_list = []

        for field in fields:
            catalogs_list.append(cat_file.format(field=field))

        ut.concat_data(files_list=catalogs_list,
                       save_file=save_file)

        ut.printlog(f"Created catalog {save_file}", log_file)

    else:
        ut.printlog(f"File {save_name} already exists", log_file)


concatenate_calib_catalogs()


# ***************************************************
#    Compute XY maps
# ***************************************************

def compute_xy_maps():

    """
    Computes XY correction maps
    """

    print("")
    ut.printlog(('********** '
                 'Computing XY correction maps '
                 '**********'),
                 log_file)
    print("")

    cat_name = 'concat_mag_ext_xycorr.cat'
    cat_file = os.path.join(save_path, cat_name)

    for filt in mag_pairs.keys():

        save_name = f"xy_corr_map_{filt}.npy"
        save_file = os.path.join(save_path, save_name)

        if not os.path.exists(save_file):

            ut.get_xy_correction_grid(data_file = cat_file,
                                      save_file = save_file,
                                      mag       = filt,
                                      mag_ref   = mag_pairs[filt],
                                      xbins     = xbins,
                                      ybins     = ybins)

            ut.printlog(f"Created file {save_file}.", log_file)

        else:
            ut.printlog(f"File {save_file} already exists.", log_file)


compute_xy_maps()


# ***************************************************
#    Plot XY maps
# ***************************************************

def plot_xy_maps():

    """
    Plots XY correction maps
    """

    print("")
    ut.printlog(('********** '
                 'Plotting XY correction maps '
                 '**********'),
                 log_file)
    print("")

    for filt in mag_pairs.keys():

        grid_name = f"xy_corr_map_{filt}.npy"
        grid_file = os.path.join(save_path, grid_name)

        save_name = f"xy_corr_map_{filt}.png"
        save_file = os.path.join(plots_path, save_name)

        if not os.path.exists(save_file):

            ut.plot_xy_correction_grid(grid_file = grid_file,
                                       save_file = save_file,
                                       mag       = filt,
                                       xbins     = xbins,
                                       ybins     = ybins)

            ut.printlog(f"Created file {save_file}.", log_file)

        else:
            ut.printlog(f"File {save_file} already exists.", log_file)


plot_xy_maps()


# -*- coding: utf-8 -*-

# ******************************************************************************
#                          S-PLUS CALIBRATION PIPELINE
#                        combine_calibration_catalogs.py
#         Combines intermediare calibration catalogs for a list of fields
# ******************************************************************************

"""
Combines intermediare calibration catalogs for a list of fields

Command line arguments for this script are, respectivelly:
1) Location of the S-PLUS fields list file
2) Location of the configuration file used in the calibration
3) Location to save the intermediary calibration catalogs

--------------------------------------------------------------------------------
   FUNCTIONS:
--------------------------------------------------------------------------------

concatenate_instrumental_catalogs()
concatenate_external_calib_catalogs()
concatenate_stlocus_calib_catalogs()
concatenate_internal_calib_catalogs()

--------------------------------------------------------------------------------
   COMMENTS:
--------------------------------------------------------------------------------
This script should be run individually and not within the pipeline.py script

--------------------------------------------------------------------------------
   USAGE:
--------------------------------------------------------------------------------
$python3 combine_calibration_catalogs.py *field_list_file* *config_file*

----------------
"""

################################################################################
# Import external packages

import os
import sys

addsteps_path = os.path.split(__file__)[0]
pipeline_path = os.path.split(addsteps_path)[0]
spluscalib_path = os.path.split(pipeline_path)[0]

sys.path.append(spluscalib_path)

################################################################################
# Import spluscalib packages

from spluscalib import utils as ut

################################################################################
# Read parameters

field_list_file = sys.argv[1]
conf_file = sys.argv[2]
save_path = sys.argv[3]

conf = ut.pipeline_conf(conf_file)

suffix = ut.calibration_suffix(conf)

################################################################################
# Initiate log file

log_file_name = os.path.join(save_path, 'combine_calibration_catalogs.log')
log_file_name = ut.gen_logfile_name(log_file_name)
log_file = os.path.join(save_path, log_file_name)

with open(log_file, "w") as log:
    log.write("")

################################################################################
# Reading field list

ut.printlog("*********** Reading field list **********", log_file)

fields = ut.load_field_list(field_list_file)

ut.printlog(f"Running the pipeline for fields:", log_file)
ut.printlog(f"{fields}", log_file)

################################################################################
# Begin script

# ***************************************************
#    Combine instrumental catalogs
# ***************************************************


def concatenate_instrumental_catalogs():

    """
    Concatenates calibration catalogs with S-PLUS instrumental magnitudes
    """

    print("")
    ut.printlog(('********** '
                 'Concatenating instrumental magnitudes '
                 '**********'),
                 log_file)
    print("")

    cat_name = '{field}_mag_inst.cat'
    cat_file = os.path.join(conf['save_path'], '{field}',
                                 f'Calibration_{suffix}', 'external',
                                 cat_name)

    # Start save name
    save_name = 'concat_mag_inst.cat'
    save_file = os.path.join(save_path, save_name)

    if not os.path.exists(save_file):

        # Create list of field catalogs
        catalogs_list = []

        for field in fields:
            catalogs_list.append(cat_file.format(field=field))

        ut.concat_data(files_list = catalogs_list,
                       save_file  = save_file)

        ut.printlog(f"Created catalog {save_file}", log_file)

    else:
        ut.printlog(f"File {save_name} already exists", log_file)


concatenate_instrumental_catalogs()


# ***************************************************
#    Combine external calibration catalogs
# ***************************************************

def concatenate_external_calib_catalogs():

    """
    Concatenates calibration catalogs after the external calibration step
    has been run
    """

    print("")
    ut.printlog(('********** '
                 'Concatenating external calibrated magnitudes '
                 '**********'),
                 log_file)
    print("")

    cat_name = '{field}_mag_ext.cat'
    cat_file = os.path.join(conf['save_path'], '{field}',
                            f'Calibration_{suffix}', 'external',
                            cat_name)

    # Start save name
    save_name = 'concat_mag_ext.cat'
    save_file = os.path.join(save_path, save_name)

    if not os.path.exists(save_file):

        # Create list of field catalogs
        catalogs_list = []

        for field in fields:
            catalogs_list.append(cat_file.format(field=field))

        ut.concat_data(files_list=catalogs_list,
                       save_file=save_file)

        ut.printlog(f"Created catalog {save_file}", log_file)

    else:
        ut.printlog(f"File {save_name} already exists", log_file)


concatenate_external_calib_catalogs()


# ***************************************************
#    Combine stellar locus calibration catalogs
# ***************************************************

def concatenate_stlocus_calib_catalogs():

    """
    Concatenates calibration catalogs after the stellar locus calibration step
    has been run
    """

    print("")
    ut.printlog(('********** '
                 'Concatenating stellar locus calibrated magnitudes '
                 '**********'),
                 log_file)
    print("")

    cat_name = '{field}_mag_stlocus.cat'
    cat_file = os.path.join(conf['save_path'], '{field}',
                            f'Calibration_{suffix}', 'stlocus',
                            cat_name)

    # Start save name
    save_name = 'concat_mag_stlocus.cat'
    save_file = os.path.join(save_path, save_name)

    if not os.path.exists(save_file):

        # Create list of field catalogs
        catalogs_list = []

        for field in fields:
            catalogs_list.append(cat_file.format(field=field))

        ut.concat_data(files_list=catalogs_list,
                       save_file=save_file)

        ut.printlog(f"Created catalog {save_file}", log_file)

    else:
        ut.printlog(f"File {save_name} already exists", log_file)


if 'calibration_stloc' in conf['run_steps']:
    concatenate_stlocus_calib_catalogs()


# ***************************************************
#    Combine internal calibration catalogs
# ***************************************************

def concatenate_internal_calib_catalogs():

    """
    Concatenates calibration catalogs after the internal calibration step
    has been run
    """

    print("")
    ut.printlog(('********** '
                 'Concatenating internal calibrated magnitudes '
                 '**********'),
                 log_file)
    print("")

    cat_name = '{field}_mag_int.cat'
    cat_file = os.path.join(conf['save_path'], '{field}',
                            f'Calibration_{suffix}', 'internal',
                            cat_name)

    # Start save name
    save_name = 'concat_mag_int.cat'
    save_file = os.path.join(save_path, save_name)

    if not os.path.exists(save_file):

        # Create list of field catalogs
        catalogs_list = []

        for field in fields:
            catalogs_list.append(cat_file.format(field=field))

        ut.concat_data(files_list=catalogs_list,
                       save_file=save_file)

        ut.printlog(f"Created catalog {save_file}", log_file)

    else:
        ut.printlog(f"File {save_name} already exists", log_file)


if 'calibration_internal' in conf['run_steps']:
    concatenate_internal_calib_catalogs()


# ***************************************************
#    Combine gaiascale calibration catalogs
# ***************************************************

def concatenate_gaiascale_calib_catalogs():

    """
    Concatenates calibration catalogs after the gaia scale calibration step
    has been run
    """

    print("")
    ut.printlog(('********** '
                 'Concatenating gaia scale calibrated magnitudes '
                 '**********'),
                 log_file)
    print("")

    cat_name = "{field}_mag_gaiascale.cat"
    cat_file = os.path.join(conf['save_path'], '{field}',
                            f'Calibration_{suffix}', 'gaiascale',
                            cat_name)

    # Start save name
    save_name = 'concat_mag_gaiascale.cat'
    save_file = os.path.join(save_path, save_name)

    if not os.path.exists(save_file):

        # Create list of field catalogs
        catalogs_list = []

        for field in fields:
            catalogs_list.append(cat_file.format(field=field))

        ut.concat_data(files_list=catalogs_list,
                       save_file=save_file)

        ut.printlog(f"Created catalog {save_file}", log_file)

    else:
        ut.printlog(f"File {save_name} already exists", log_file)


if 'calibration_gaiascale' in conf['run_steps']:
    concatenate_gaiascale_calib_catalogs()


# ***************************************************
#    Combine diagnostic calibration catalogs
# ***************************************************

def concatenate_diagnostic_calib_catalogs():

    """
    Concatenates calibration catalogs after the diagnostic calibration step
    has been run
    """

    print("")
    ut.printlog(('********** '
                 'Concatenating diagnostic magnitudes '
                 '**********'),
                 log_file)
    print("")

    cat_name = '{field}_diagnostic_mag.fits'
    cat_file = os.path.join(conf['save_path'], '{field}',
                            f'Calibration_{suffix}', 'diagnostic',
                            cat_name)

    # Start save name
    save_name = 'concat_mag_diagnostic.cat'
    save_file = os.path.join(save_path, save_name)

    if not os.path.exists(save_file):

        # Create list of field catalogs
        catalogs_list = []

        for field in fields:
            catalogs_list.append(cat_file.format(field=field))

        ut.concat_data(files_list=catalogs_list,
                       save_file=save_file)

        ut.printlog(f"Created catalog {save_file}", log_file)

    else:
        ut.printlog(f"File {save_name} already exists", log_file)


if 'calibration_diagnostic' in conf['run_steps']:
    concatenate_diagnostic_calib_catalogs()



#====================================================================================================================#

# For general purposes
import numpy as np

# To message the user
import sys

# To deal with the data as a dataframe
import pandas as pd

# To calculate tr priors
from scipy.stats import gaussian_kde as kde

# Just for type hinting
from typing import Callable

#====================================================================================================================#


class PriorCalculator:

	"""
    This class is used as a calculator of prior probabilities of models using KDEs, it is implemented calculating
	nested probabilities for que parameters passed in, for example, if tuples of the parameters (p0, p1, p2), are
	passed inthe priors returned are simply P(q0) * P(q1 | q0) * P(q2 | q1, q0), where each of this probabilities
	is calculated with a KDE inside bins of previous parameters, in this example, P(q0) is just the value of q0 in
	the total KDE of p0, but P(q1 | q0) is the value of q1 in the KDE calculated inside a bin p0, the same for
	P(q2 | q1, q0), it's calculated inside a bin of p0 and p1. (The normalization method is up to the user to choose)

    In the contructor, the calculator receives the bins that you want to use on the calculation and a bandwidth_factor
	(bw) to be used on the KDEs. It's optional to the user to instead of passing in the bins, constructing them via
	the method self.regular_bins(...) which, as the name sugests, creates N bins of equal size in the given	range for
	every parameter.
	
	Methods:
    ------------
    
    """
	
	def __init__(self, bins = [], bw = 'scott', normalize_by = 'max'):
		
		self.bw = bw
		self.bins = bins
		self.normalize_by = normalize_by


	def regular_bins(self, vectors):

		'''
		This is meant to be called when the choice of bins is such that the 
		limits are evenly spaced. (It constructs the bins, therefore, only use
		when the bins are not passed in the constructor).

		Parameters:
    	------------
		vectors: list
			A list of ordered tuples of the kind (N, r0, r1) where N is the
			number of bins, r0 is the lower limit and r1 is the upper limit.
			(If there is m parameters, vectors should contain m - 1 tuples).
		'''

		# For every parameter (except for the last)
		for N, r0, r1 in vectors:
			
			# Use the boundaries and N to define the separation
			step = (r1 - r0) / N

			# Constructs the bins for that parameter
			bins = [(r0 + i * step, r0 + (i + 1) * step)
					 for i in range(N)]
			
			# Adds it to the bin list
			self.bins.append(bins)


	def interpret(self, ref, cols, fmt, **kwargs):

		'''
		This method takes the file passed as a reference, reads it
		into a dataframe and than fills a tensor of KDE vectors for
		each bin.

		Parameters:
    	------------
		ref: str
			Path to the reference that's going to be used as a base
			for the prior calculation (as by using the density as a
			probability).
		cols: list[str]
			Name of the columns with the parameters (in order) that
			are going to be used.
		fmt: str
			Format of the file passed in the ref argument.
		'''
		
		# If it is a .cat file (ascii)
		if (fmt == 'ascii'):

			# Opens the file into a dataframe
			self.ref = pd.read_csv(ref,
								   delim_whitespace = True,
								   escapechar       = '#',
							       skipinitialspace = True)
			self.ref.columns = self.ref.columns.str.replace(' ', '')

		# Populates a tensor with all the KDEs
		self.fill_bins(cols, **kwargs)


	def indexes(self, theta):

		'''
		This method is used to find the indexes of the bins that a
		vector (theta) fits inside.

		Parameters:
    	------------
		theta: tuple
			Parameter vector (p0, ..., pn).

		Returns:
		------------
		indexes: tuple
			Indexes of the bin that contains theta.
		'''

		# Initializes the index list with -1s 
		indexes = [ -1 for _ in self.bins]

		# For each parameter (except for the last)
		for p in range(len(self.bins)):

			# Goes through the bins (if the bin is not found it keeps the -1)
			for i, b in enumerate(self.bins[p]):

				b0, b1 = b

				# Until it finds the bin that contains the parameter
				if (theta[p] >= b0) and (theta[p] < b1):

					# it adds the index to the list
					indexes[p] = i

					# and goes to the next parameter
					break
				
		return tuple(indexes)


	def fill_bins(self, cols, lastBound = [-10000, 10000]):
		
		'''
		This method is used to fill the self.prior_bins tensor with KDE vectors that
		are going to be accessed when calculating the priors.

		(This is a generalization of the process of calculating KDEs inside bins of 
		previous parameters.)

		How it works:
		------------

			The cells (c[0], ..., c[n-1]) have two properties, a controlling index
			(STATE) and a list (STORAGE), they are used to construct the tensor that
			contains the KDE vectors.

			A cell is said to be FULL when it's STATE reaches the length of the bin set
			of the corresponding parameter.

			The local reference array is defined in such a way that the element i is
			a restricted version of the element in i - 1 (in practice, every new element
			is the previous but reduced to the data contained on the current bin).

			We first shall initialize cells with STATE = 0 and empty STORAGE (= []),
			one for each parameter that has bins.

			> Start a KDE vector with the kde for p0 (which is independent of bins) 

			> For every other parameter, starting from p1, with the last local reference,
			take it's intersection with the current bin of the previous parameter. Add this
			new reference to the local references and use it to calculate the KDE for
			the current parameter. Add the KDE to the vector.

			> Add the complete KDE vector to the STORAGE of the last cell.

			> For every cell, biginning from the last, if the cell is FULL, set it's
			STATE back to 0 and put it's STORAGE inside the STORAGE of the cell to it's left.

			> If the cell is not FUll yet, up it's STATE by 1;

			When cell[0] is FULL the program is over, the tensor is it's STORAGE.

		Parameters:
    	------------
		cols: list[str]
			Name of the columns with the parameters (in order) that
			are going to be used.
		'''

		# Sorts the bins (just in case)
		for b in self.bins:
			b.sort()

		# Initializes the cells with STATE = 0 adn empty STORAGE
		cells = [[0, [ ]] for _ in self.bins]

		# Do while there's still bins of p0 to cover
		while cells[0][0] < len(self.bins[0]):

			# Message to the user
			message = f"Populating bins with KDEs... indexes = \u001b[38;5;219m{[c[0] for c in cells]}\u001b[0m"
			sys.stdout.write(f'\u001b[32;1m>>>\u001b[0m {message} \r')
			sys.stdout.flush()

			# Initializes the local_refs array with the global reference in it
			local_refs  = [self.ref]

			# Calculates the KDE for p0
			K0 = kde(local_refs[-1][cols[0]], bw_method = self.bw)
			
			# Initializes an empty KDE vector
			KDEs = [K0]

			# For every parameter (that has bins) (except for p0)
			for i in range(1, len(cols)):

				# Takes the instersection between the last local reference and the current bin
				lower = local_refs[-1][cols[i - 1]] >= self.bins[i - 1][cells[i - 1][0]][0]
				upper = local_refs[-1][cols[i - 1]] <  self.bins[i - 1][cells[i - 1][0]][1]

				# And adds it to local references
				local_refs.append(local_refs[-1].loc[lower & upper])
				
				# Crates the KDE inside all the previous bins.
				bounds = lastBound if (i + 1 == len(cols)) else [self.bins[i][0][0], self.bins[i][-1][-1]]
				Ki = self.normKDE(kde(np.array(local_refs[-1][cols[i]]), bw_method = self.bw), bounds)
				
				# Appends the KDE to the vector
				KDEs.append(Ki)
		
			# Appends the vector to the STORAGE of the last cell
			cells[-1][1].append(KDEs)

			# Updates the cells
			J = -1
			while J + len(self.bins) >= 0:

				# If the current cell is FULL
				if (cells[J][0] + 1 == len(self.bins[J])):

					# If the cell is c[0] than it's all finished
					if (J + len(self.bins) == 0):
						cells[J][0] += 1
						break
					
					# If not, puts a copy of current STORAGE inside the STORAGE to the left
					cells[J - 1][1].append(cells[J][1][:])
					
					# Resets the current cell
					cells[J] = [0, [ ]]

					# Update J
					J -= 1
				
				# If the current cell is not yet Full
				else:

					# Just update it's STATE
					cells[J][0] += 1
					break

				
		# When finished, the tensor is just the STORAGE of cell[0]
		self.prior_bins = np.array(cells[0][1], dtype=object)


	def calculate(self, theta):

		'''
		For the given vector of parameters (p0, p1, ..., pn), calculates the prior
		probability (still not normalized) of that vector via the equation

			P(p) = P(p0) P(p1 | p0) P(p2 | p0, p1) ... P(pn | p0, ..., p(n-1)).

		In practical terms it means that it finds the bins that contain the vector
		and multiply the values obtained from the KDEs inside the bins.

		Parameters:
    	------------
		theta: tuple
			Vector with the parameters that we want to calculate the prior to.

		Returns:
		------------
		prior: float
			The not still normalized calculated prior.
		'''
		
		# Gets the indexes of the bins that contain theta
		indexes = self.indexes(theta)

		# If one of the indexes is -1, by our convention it means theta is out of the boudaries
		if (-1 in indexes):

			# Therefore should be assigned a value of 0 to prior
			return 0

		# Access the tensor at that indexes to get the corresponding KDE vector
		vec = self.prior_bins[indexes]

		# Initializes the prior at 1
		prior = 1

		# For every KDE in the vector
		for i, KDE in enumerate(vec):

			# Calculates the value of the parameter on the KDE and multiplies the prior by it
			prior *= KDE(theta[i])[0]

		return prior
	

	def prior_column(self, parameters):

		'''
		For a given list of parameter vectors, calculates the prior for each 
		one and returns it into a list.

		Parameters:
    	------------
		parameters: list[tuple]
			List of parameter tuples for which this function calculates
			the prior to.
		
		Returns:
		------------
		normalized_column: list
			An ordered list representing a column with the normalize priors
			of the parameter tuples passed as input.
		'''

		# Creates the column with the prior For every parameter vector
		column = [self.calculate(theta) for theta in parameters]

		# Creates a normalized version of the column
		normalized_column = self.normalized(column, by = self.normalize_by)

		return normalized_column


	def normalized(self, column, by):
		
		'''
		This method normalizes the prior values in a given column using
		the method specified bu the user (by sum, by max, ...).

		Parameters:
    	------------
		column: list[float]
			Column containing the prior values to be normalized.
		by: str
			Method to be used in the normalization.

		Returns:
		------------
		normalized_column: list
			The same column as the input but with normalized values.
		'''

		# If the method is a normalization by maximum
		if (by == "max"):

			# The normalization factor is just the biggest value
			factor = max(column)

		# If the method is a normalization by sum
		elif (by == "sum"):

			# We should use the sum of the priors in the column as a factor
			factor = sum(column)
		
		else:
			return column

		# Returns the list with all the values divided by the calculated factor
		normalized_column = list(map(lambda x: x / factor, column))
		return normalized_column
	

	def normKDE(self, KDE, bounds):

		'''
		Given a gaussian_kde returns a normalized version of it.

		Parameters:
    	------------
		KDE: Callable[[float], float]
			Kernel Density Estimator to be normalized.
		bounds: list[gloat]
			KDE boundaries (to calculate the integral).

		Returns:
		------------
		nKDE: Callable[[float], float]:
			Normalized KDE.
		'''

		area = KDE.integrate_box(*bounds)
		nKDE = (lambda x: KDE(x) / area)
		return nKDE


# -*- coding: utf-8 -*-

# ******************************************************************************
#                          S-PLUS CALIBRATION PIPELINE
#                            calibration_finalzp.py
#      Combines the zero-points of all calibration steps into a final file
# ******************************************************************************

"""
Combines the zero-points of all calibration steps into a final zp file

The S-PLUS field is given as the first command line argument. Configurations
are set in the config file, given as the second command line argument.

--------------------------------------------------------------------------------
   FUNCTIONS:
--------------------------------------------------------------------------------
copy_offset_zp_file()
calculate_final_zp()

--------------------------------------------------------------------------------
   COMMENTS:
--------------------------------------------------------------------------------
Ideally this script should only be run through the pipeline.py script.

Assumes that at least calibration_external.py, and possibly calibration_stloc.py
and calibration_internal.py, have already been run for this field.

--------------------------------------------------------------------------------
   USAGE:
--------------------------------------------------------------------------------
$python3 calibration_finalzp.py *field_name* *config_file*

----------------
"""

################################################################################
# Import external packages

import os
import sys

steps_path = os.path.split(__file__)[0]
pipeline_path = os.path.split(steps_path)[0]
spluscalib_path = os.path.split(pipeline_path)[0]

sys.path.append(spluscalib_path)

################################################################################
# Import spluscalib packages

from spluscalib import utils as ut

################################################################################
# Read parameters

field     = sys.argv[1]
conf_file = sys.argv[2]

conf = ut.pipeline_conf(conf_file)

################################################################################
# Get directories

field_path       = os.path.join(conf['run_path'], field)
crossmatch_path  = os.path.join(field_path, 'Crossmatch')

suffix = ut.calibration_suffix(conf)
calibration_path  = os.path.join(field_path, f'Calibration_{suffix}')

external_cal_path = os.path.join(calibration_path, 'external')
external_plot_path = os.path.join(external_cal_path, 'plots')

stloc_cal_path = os.path.join(calibration_path, 'stlocus')
stloc_plot_path = os.path.join(stloc_cal_path, 'plots')

internal_cal_path = os.path.join(calibration_path, 'internal')
internal_plot_path = os.path.join(internal_cal_path, 'plots')

gaiascale_cal_path  = os.path.join(calibration_path, 'gaiascale')
gaiascale_plot_path = os.path.join(gaiascale_cal_path, 'plots')


log_path = os.path.join(calibration_path, 'logs')


################################################################################
# Initiate log file

ut.makedir(calibration_path)
ut.makedir(internal_cal_path)
ut.makedir(internal_plot_path)
ut.makedir(log_path)

log_file_name = os.path.join(log_path, 'calibration_finalzp.log')
log_file_name = ut.gen_logfile_name(log_file_name)
log_file = os.path.join(calibration_path, log_file_name)

with open(log_file, "w") as log:
    log.write("")


################################################################################
# Begin script

# ***************************************************
#    Combine all zps
# ***************************************************

def calculate_final_zp():
    """
    Combines all calibration .zp files to create a final zp file for the field
    """

    print("")
    ut.printlog(('********** '
                 'Combining all calibration .zp files '
                 '**********'),
                log_file)
    print("")

    zp_file_list = []

    save_name = f"{field}_final.zp"
    save_file = os.path.join(calibration_path, save_name)

    if not os.path.exists(save_file):
        # Add external .zp
        ext_zp_name = f"{field}_external.zp"
        ext_zp_file = os.path.join(external_cal_path, ext_zp_name)

        zp_file_list.append(ext_zp_file)

        # Add stellar locus .zp
        if 'calibration_stloc' in conf['run_steps']:
            stloc_zp_name = f"{field}_stlocus.zp"
            stloc_zp_file = os.path.join(stloc_cal_path, stloc_zp_name)

            zp_file_list.append(stloc_zp_file)

        # Add internal .zp
        if 'calibration_internal' in conf['run_steps']:
            int_zp_name = f"{field}_internal.zp"
            int_zp_file = os.path.join(internal_cal_path, int_zp_name)

            zp_file_list.append(int_zp_file)

        # Add gaia scale .zp
        if 'calibration_gaiascale' in conf['run_steps']:
            gaiascale_zp_name = f"{field}_gaiascale_splus.zp"
            gaiascale_zp_file = os.path.join(gaiascale_cal_path,
                                             gaiascale_zp_name)

            zp_file_list.append(gaiascale_zp_file)

        filters = []
        for filt in conf['filters']:
            filters.append(f"SPLUS_{filt}")

        # Combine zero-points
        ut.zp_add(zp_file_list = zp_file_list,
                  save_file    = save_file,
                  filters      = filters,
                  inst_zp      = conf['inst_zp'])

        ut.printlog(f"Created file {save_file}", log_file)

    else:
        ut.printlog(f"File {save_name} already exists", log_file)


calculate_final_zp()



# -*- coding: utf-8 -*-

# ******************************************************************************
#                          S-PLUS CALIBRATION PIPELINE
#                              photometry_dual.py
#                       aperture photometry in dual mode
# ******************************************************************************

"""
Generates a detection image and runs SExtractor in dual mode for one S-PLUS
field and a list of filters.

The S-PLUS field is given as the first command line argument.

The set of filters, location of S-PLUS images, default SExtractor configuration
and param files, filters used to build the detection image, and the output path
must be set in the configuration file given as the second command line argument.

--------------------------------------------------------------------------------
   FUNCTIONS:
--------------------------------------------------------------------------------

create_photometry_path()
generate_swarp_config()
generate_combine_images_list()
generate_detection_image()
update_detection_header()
generate_detection_sexconfig_file()
generate_dual_sexconfig_files()
run_sex_detection()
run_sex_dual()
convert_detection_fits_to_fz()
dual_diagnostic_plots()

--------------------------------------------------------------------------------
   COMMENTS:
--------------------------------------------------------------------------------
Ideally this script should only be run through the pipeline.py script.

--------------------------------------------------------------------------------
   USAGE:
--------------------------------------------------------------------------------
$python3 photometry_dual.py *field_name* *config_file*

----------------
"""

################################################################################
# Import external packages

import os
import sys

steps_path = os.path.split(__file__)[0]
pipeline_path = os.path.split(steps_path)[0]
spluscalib_path = os.path.split(pipeline_path)[0]

sys.path.append(spluscalib_path)

################################################################################
# Import spluscalib packages

from spluscalib import utils as ut

################################################################################
# Read parameters

field     = sys.argv[1]
conf_file = sys.argv[2]

conf = ut.pipeline_conf(conf_file)

################################################################################
# Get directories

field_path      = os.path.join(conf['run_path'], field)
photometry_path = os.path.join(field_path, 'Photometry')
dual_path       = os.path.join(photometry_path, 'dual')
sexconf_path    = os.path.join(dual_path, 'sexconf')
catalogs_path   = os.path.join(dual_path, 'catalogs')
detection_path  = os.path.join(dual_path, 'detection')

images_path     = os.path.join(field_path, 'Images')

log_path        = os.path.join(photometry_path, 'logs')

################################################################################
# Create Photometry directory

ut.makedir(field_path)
ut.makedir(photometry_path)
ut.makedir(log_path)

################################################################################
# Initiate log file

log_file_name = os.path.join(log_path, 'photometry_dual.log')
log_file_name = ut.gen_logfile_name(log_file_name)
log_file = os.path.join(photometry_path, log_file_name)

with open(log_file, "w") as log:
    log.write("")

################################################################################
# Log configuration

ut.printlog("Dual mode photometry parameters:", log_file,
            color = "blue", attrs=["bold"])

dual_mode_params = ['run_path', 'filters', 'path_to_images', 'inst_zp',
                    'use_weight', 'path_to_sex', 'sex_config', 'sex_param',
                    'detection_image']

for param in dual_mode_params:
    try:
        ut.printlog(f"{param}: {conf[param]}", log_file,
                    color = "blue")
    except KeyError:
        ut.printlog(f"{param}: NONE", log_file,
                     color = "red")

################################################################################
# Begin script

# ***************************************************
#    If photometry path does not exist -> create
# ***************************************************


def create_photometry_path():

    print("")
    ut.printlog(('********** '
                 'Generating Dual Photometry paths '
                 '**********'),
                 log_file, color = "yellow")
    print("")

    # Create dual mode photometry path
    ut.makedir(dual_path)

    # Create dual mode sexconf path
    ut.makedir(sexconf_path)

    # Create dual mode catalogs path
    ut.makedir(catalogs_path)

    # Create detection image path
    ut.makedir(detection_path)


create_photometry_path()


# ***************************************************
#  Generate SWARP configuration for detection image
# ***************************************************

def generate_swarp_config():
    """
    Generate swarp configuration file for detection image
    """

    print("")
    ut.printlog(('********** '
                 'Generating swarp configuration '
                 '**********'), log_file, color = "yellow")
    print("")

    save_file = os.path.join(detection_path, f'{field}_config.swarp')

    if not os.path.exists(save_file):

        detection_image_name = f'{field}_detection.fits'
        detection_weight_name = f'{field}_detectionweight.fits'

        detection_image_file = os.path.join(detection_path,
                                            detection_image_name)
        detection_weight_file = os.path.join(detection_path,
                                             detection_weight_name)
        xml_output_file = os.path.join(detection_path, 'swarp_output.xml')

        if conf['use_weight']:
            combine_type = "WEIGHTED"
            weight_type  = "MAP_WEIGHT"
        else:
            combine_type = "AVERAGE"
            weight_type  = "NONE"

        # Get first image for center reference
        ref_filt = conf['detection_image'][0]
        image_name = f"{field}_{ref_filt}_swp.fits"
        image_file = os.path.join(images_path, image_name)

        # Generating config file
        ut.get_swarp_config(save_file            = save_file,
                            default_swarpconfig  = conf['swarp_config'],
                            detection_image_out  = detection_image_file,
                            detection_weight_out = detection_weight_file,
                            resample_dir         = detection_path,
                            xml_output           = xml_output_file,
                            combine_type         = combine_type,
                            weight_type          = weight_type,
                            ref_image            = image_file)

        ut.printlog(f"Created file {save_file}", log_file,
                    color = "green")

    else:
        ut.printlog(f"File {save_file} already created.", log_file,
                    color = "white", attrs=["dark"])


generate_swarp_config()


# ***************************************************
#  Generate combine images list
# ***************************************************

def generate_combine_images_list():

    """
    Generate list of images to combine for the detection image
    """

    print("")
    ut.printlog(('********** '
                 'Generating list of images to combine '
                 '**********'), log_file, color = "yellow")
    print("")

    save_file = os.path.join(detection_path, f'combine_images.txt')

    if not os.path.exists(save_file):
        # Generating filter image list
        with open(save_file, 'w') as imlist:
            for filt in conf['detection_image']:
                # Get image file
                image_name = f"{field}_{filt}_swp.fits"
                image_file = os.path.join(images_path, image_name)

                imlist.write(image_file)
                imlist.write("\n")

        ut.printlog(f"Created file combine_images.txt", log_file,
                    color = "green")

    else:
        ut.printlog(f"{save_file} already created.", log_file,
                    color = "white", attrs = ["dark"])


generate_combine_images_list()

# ***************************************************
#    Generate Detection image
# ***************************************************


def generate_detection_image():

    """
    Generate detection image
    """

    print("")
    ut.printlog(('********** '
                 'Generating Detection Image '
                 '**********'), log_file, color = "yellow")
    print("")

    det_image_name = f"{field}_detection.fits"
    det_image_file = os.path.join(detection_path, det_image_name)
    det_image_fz   = det_image_file + ".fz"

    if not os.path.exists(det_image_fz):
        if not os.path.exists(det_image_file):

            # Get swarp config file name
            swarp_conf = os.path.join(detection_path, f'{field}_config.swarp')

            # Get combine images list
            image_list = os.path.join(detection_path, 'combine_images.txt')

            # Prepare swarp cmd
            cmd = f"{conf['path_to_swarp']} @{image_list} -c {swarp_conf}"

            ut.printlog("$ "+cmd, log_file, color = "cyan")

            # Run swarp
            os.system(cmd)

        else:
            ut.printlog(f"File {det_image_file} already exists.", log_file,
                        color = white, attrs=["dark"])
    else:
        ut.printlog(f"File {det_image_fz} already exists.", log_file,
                    color = "white", attrs=["dark"])


generate_detection_image()


# ***************************************************
#    Update Detection Image Header
# ***************************************************

def update_detection_header():

    """
    Update detection image header
    """

    print("")
    ut.printlog(('********** '
                 'Updating detection image header '
                 '**********'),
                log_file, color = "yellow")
    print("")

    # Get detection image file
    det_image_name = f"{field}_detection.fits"
    det_image_file = os.path.join(detection_path, det_image_name)

    # Get list of comined images
    image_list = os.path.join(detection_path, f'combine_images.txt')

    # Create header update control file
    header_log = os.path.join(detection_path, 'updated_header.txt')

    if not os.path.exists(header_log):

        ut.update_detection_header(detection_image = det_image_file,
                                   image_list_file = image_list)

        ut.printlog("Detection header was updated.", log_file)

        with open(header_log, 'w') as f:
            f.write("Detection header was updated.")

    else:
        ut.printlog("Detection header already updated.", log_file,
                    color = "white", attrs=["dark"])


update_detection_header()

# ***************************************************
#    Generate Detection SExconfig file
# ***************************************************


def generate_detection_sexconfig_file():

    """
    Generate detection image sexconfig file
    """

    print("")
    ut.printlog(('********** '
                 'Generating detection image photometry config files '
                 '**********'),
                log_file, color = "yellow")
    print("")

    # Default SExconfig and SExparam files
    default_sexconfig = conf['sex_config']
    default_sexparam = conf['sex_param']

    conf_file_name = f"{field}_detection.sex"
    save_file = os.path.join(detection_path, conf_file_name)

    # Generate location of output catalog file
    catalog_name = f"sex_{field}_detection_catalog.fits"
    catalog_file = os.path.join(detection_path, catalog_name)

    # Get location of image file
    image_name = f"{field}_detection.fits"
    image_file = os.path.join(detection_path, image_name)

    segm_image = image_file.replace(".fits", ".segm.fits")
    aper_image = image_file.replace(".fits", ".aper.fits")

    if not os.path.exists(save_file):

        ut.get_sex_config(save_file=save_file,
                          default_sexconfig=default_sexconfig,
                          default_sexparam=default_sexparam,
                          catalog_file=catalog_file,
                          image_file=image_file,
                          inst_zp=conf['inst_zp'],
                          path_to_sex=conf['path_to_sex'],
                          use_weight=conf['use_weight'],
                          mode='single',
                          check_segima=segm_image,
                          check_aperima=aper_image)

        ut.printlog(f"Generated detection image sexconfig for field {field}",
                    log_file, color = "green")

    else:
        ut.printlog(("Detection image sexconfig already obtained for field "
                     f"{field}"), log_file, color = "white", attrs=["dark"])


generate_detection_sexconfig_file()

# ***************************************************
#    Generate SExconfig files
# ***************************************************


def generate_dual_sexconfig_files():

    """
    Generate dual mode sexconfig file for all filters
    """

    print("")
    ut.printlog(('********** '
                 'Generating Dual Photometry config files '
                 '**********'),
                 log_file, color = "yellow")
    print("")

    # Default SExconfig and SExparam files
    default_sexconfig = conf['sex_config']
    default_sexparam  = conf['sex_param']

    # Generate configuration for each filter
    for filt in conf['filters']:

        # Generate location of config file
        conf_file_name = f"{field}_{filt}_dual.sex"
        save_file = os.path.join(sexconf_path, conf_file_name)

        # Generate location of output catalog file
        catalog_name = f"sex_{field}_{filt}_dual.fits"
        catalog_file = os.path.join(catalogs_path, catalog_name)

        # Get location of image file
        image_name = f"{field}_{filt}_swp.fits"
        image_file = os.path.join(images_path, image_name)

        # Get location of detection image file
        detection_name = f"{field}_detection.fits"
        detection_file = os.path.join(detection_path, detection_name)

        if not os.path.exists(save_file):

            ut.get_sex_config(save_file = save_file,
                              default_sexconfig = default_sexconfig,
                              default_sexparam  = default_sexparam,
                              catalog_file      = catalog_file,
                              image_file        = image_file,
                              inst_zp           = conf['inst_zp'],
                              path_to_sex       = conf['path_to_sex'],
                              use_weight        = conf['use_weight'],
                              mode              = 'dual',
                              detection_file    = detection_file)

            ut.printlog(("Generated dual mode sexconfig for field "
                         f"{field}, filter {filt}"), log_file,
                         color = "green")

        else:
            ut.printlog(("Dual mode sexconfig already obtained for field "
                         f"{field}, filter {filt}"), log_file,
                         color = "white", attrs=["dark"])

    # Copy sexparam file to field config directory (for proper documentation)
    cmd = f'cp {default_sexparam} {sexconf_path}'
    ut.printlog("$ "+cmd, log_file, color = "green")
    os.system(cmd)


generate_dual_sexconfig_files()

# ***************************************************
#    Run SExtractor in Detection image
# ***************************************************


def run_sex_detection():

    """
    Run SExtractor in Detection image
    """

    print("")
    ut.printlog(('********** '
                 'Running SeX on detection image '
                 '**********'),
                 log_file, color = "yellow")
    print("")

    # Generate location of output catalog file
    catalog_name = f"sex_{field}_detection_catalog.fits"
    catalog_file = os.path.join(detection_path, catalog_name)

    if not os.path.exists(catalog_file):

        # Get location of image file
        image_name = f"{field}_detection.fits"
        image_file = os.path.join(detection_path, image_name)

        # Get location of sexconfig file
        conf_file_name = f"{field}_detection.sex"
        sexconf_file   = os.path.join(detection_path, conf_file_name)

        # Run SExtractor
        sex_location = os.path.join(conf['path_to_sex'], 'src', 'sex')

        cmd = f"{sex_location} {image_file} -c {sexconf_file}"
        cmd += " -PHOT_APERTURES '5.45455, 36.363636'"
        ut.printlog("$ "+cmd, log_file, color = "cyan")
        os.system(cmd)

    else:
        ut.printlog(("Detection image photometry already obtained for field "
                     f"{field}"), log_file, color = "white", attrs = ["dark"])


run_sex_detection()


# ***************************************************
#    Run SExtractor in Dual Mode
# ***************************************************

def run_sex_dual():

    """
    Run SExtractor in Dual mode for all filters
    """

    print("")
    ut.printlog(('********** '
                 'Running SeX in dual mode '
                 '**********'),
                 log_file, color = "yellow")
    print("")

    for filt in conf['filters']:

        # Generate location of output catalog file
        catalog_name = f"sex_{field}_{filt}_dual.fits"
        catalog_file = os.path.join(catalogs_path, catalog_name)

        if not os.path.exists(catalog_file):

            # Get location of image file
            image_name = f"{field}_{filt}_swp.fits"
            image_file = os.path.join(images_path, image_name)

            # Get location of detection image file
            detection_name = f"{field}_detection.fits"
            detection_file = os.path.join(detection_path, detection_name)

            # Generate location of config file
            conf_file_name = f"{field}_{filt}_dual.sex"
            sexconf_file   = os.path.join(sexconf_path, conf_file_name)

            # Run SExtractor
            sex_location = os.path.join(conf['path_to_sex'], 'src', 'sex')

            cmd = (f"{sex_location} {detection_file}, {image_file} "
                   f"-c {sexconf_file}")

            ut.printlog("$ "+cmd, log_file, color = "cyan")
            os.system(cmd)

        else:
            ut.printlog(("Dual mode photometry already obtained for field "
                         f"{field}, filter {filt}"), log_file,
                         color = "white", attrs = ["dark"])


run_sex_dual()

# ***************************************************
#    Convert detection .fits to .fz
# ***************************************************


def convert_detection_fits_to_fz():

    """
    Converts detection image, detection weight, detection .aper and
    detection .segm from .fits to .fz

    Then, remove the .fits files.
    """

    print("")
    ut.printlog(('********** '
                 'Converting detection .fits to .fz '
                 '**********'),
                 log_file, color = "yellow")
    print("")

    det_image_name = f"{field}_detection.fits"
    det_image_file = os.path.join(detection_path, det_image_name)

    det_weight_file = det_image_file.replace(".fits", "weight.fits")
    det_aper_file   = det_image_file.replace(".fits", ".aper.fits")
    det_segm_file   = det_image_file.replace(".fits", ".segm.fits")

    images = [det_image_file, det_weight_file, det_aper_file, det_segm_file]
    names  = ['Detection', 'Detection weight', 'Detection aper',
              'Detection segm']

    for image, name in zip(images, names):
        if not os.path.exists(image+'.fz'):

            # Compress fits
            cmd = f"fpack {image}"
            ut.printlog("$ " + cmd, log_file, color = "cyan")
            os.system(cmd)

        else:
            ut.printlog(f"{name} image already compressed.", log_file,
                        color = "white", attrs = ["dark"])

        if os.path.exists(image + '.fz'):
            # Remove fits
            cmd = f"rm {image}"
            ut.printlog("$ " + cmd, log_file, color = "red")
            os.system(cmd)

        else:
            ut.printlog(f"{name} fits image already deleted.", log_file,
                        color = "white", attrs = ["dark"])


convert_detection_fits_to_fz()

# ***************************************************
#    Make diagnostic plots
# ***************************************************

def dual_diagnostic_plots():

    """
    Makes diagnostic plots of the resulting catalogs
    """

    print("")
    ut.printlog(('********** '
                 'Making diagnostic plots for dual mode photometry'
                 '**********'),
                log_file, color = "yellow")
    print("")

    for filt in conf['filters']:

        # Generate location of output catalog file
        catalog_name = f"sex_{field}_{filt}_dual.fits"
        catalog_file = os.path.join(catalogs_path, catalog_name)

        save_name = f"{field}_{filt}_dual.png"
        save_file = os.path.join(catalogs_path, save_name)

        conf_file_name = f"{field}_{filt}_dual.sex"
        sexconf_file = os.path.join(sexconf_path, conf_file_name)

        if not os.path.exists(save_file):

            ut.plot_sex_diagnostic(catalog   = catalog_file,
                                   save_file = save_file,
                                   s2ncut    = conf['apercorr_s2ncut'],
                                   starcut   = conf['apercorr_starcut'],
                                   sexconf   = sexconf_file,
                                   filt      = filt)

            ut.printlog(f"Created file {save_file}", log_file,
                        color = "green")

        else:
            ut.printlog(("Diagnostic plots already made for field "
                         f"{field}, filter {filt}"), log_file,
                         color = "white", attrs = ["dark"])


dual_diagnostic_plots()


ut.printlog("COMPLETED DUAL MODE PHOTOMETRY", log_file, 
        color = "yellow", attrs=["bold", "reverse"])



# -*- coding: utf-8 -*-

# ******************************************************************************
#                          S-PLUS CALIBRATION PIPELINE
#                             photometry_master.py
#                      Generate photometry master catalogs
# ******************************************************************************

"""
Generates a master catalog combining the photometry of each individual field,
for each photometry mode (SExtractor single and dual modes, and dophot PSF)

The S-PLUS field is given as the first command line argument.

The set of filters, location of S-PLUS images, default SExtractor configuration
and param files, and the output path must be set in the configuration file
given as the second command line argument

--------------------------------------------------------------------------------
   FUNCTIONS:
--------------------------------------------------------------------------------

create_master_path()
extract_photometry_only_dual_catalog()
create_master_photometry_dual()
format_photometry_master_dual()
extract_photometry_only_single_catalog()
create_master_photometry_single()
format_photometry_master_single()
extract_photometry_only_psf_catalog()
create_master_photometry_psf()
format_photometry_master_psf()

--------------------------------------------------------------------------------
   COMMENTS:
--------------------------------------------------------------------------------
Ideally this script should only be run through the pipeline.py script.

Assumes that photometry_psf.py and/or photometry_single.py and/or
photometry_dual.py (and optionally correction_xy.py) and correction_aper has
already been run for this field.

--------------------------------------------------------------------------------
   USAGE:
--------------------------------------------------------------------------------
$python3 photometry_master.py *field_name* *config_file*

----------------
"""

################################################################################
# Import external packages

import os
import sys

steps_path = os.path.split(__file__)[0]
pipeline_path = os.path.split(steps_path)[0]
spluscalib_path = os.path.split(pipeline_path)[0]

sys.path.append(spluscalib_path)

################################################################################
# Import spluscalib packages

from spluscalib import utils as ut

################################################################################
# Read parameters

field     = sys.argv[1]
conf_file = sys.argv[2]

conf = ut.pipeline_conf(conf_file)


################################################################################
# Get directories

field_path      = os.path.join(conf['run_path'], field)
photometry_path = os.path.join(field_path, 'Photometry')

single_path         = os.path.join(photometry_path, 'single')
single_catalog_path = os.path.join(single_path, 'aper_correction')
single_master_path  = os.path.join(single_path, 'master')

dual_path         = os.path.join(photometry_path, 'dual')
dual_catalog_path = os.path.join(dual_path, 'aper_correction')
dual_master_path  = os.path.join(dual_path, 'master')

psf_path         = os.path.join(photometry_path, 'psf')
psf_catalog_path = os.path.join(psf_path, 'catalogs')
psf_xycorrection_path = os.path.join(psf_path, 'xy_correction')
psf_master_path  = os.path.join(psf_path, 'master')

images_path     = os.path.join(field_path, 'Images')

log_path        = os.path.join(photometry_path, 'logs')


################################################################################
# Initiate log file

ut.makedir(log_path)

log_file_name = os.path.join(log_path, 'photometry_master.log')
log_file_name = ut.gen_logfile_name(log_file_name)
log_file = os.path.join(photometry_path, log_file_name)

with open(log_file, "w") as log:
    log.write("")

################################################################################
# Log configuration

ut.printlog("Single mode photometry parameters:", log_file)

master_phot_params = ['run_path', 'filters', 'path_to_images', 'inst_zp',
                      'use_weight', 'path_to_sex', 'sex_config', 'sex_param',
                      'apercorr_diameter', 'apercorr_s2ncut',
                      'apercorr_starcut', 'apercorr_max_aperture',
                      'path_to_dophot', 'dophot_config']

for param in master_phot_params:
    try:
        ut.printlog(f"{param}: {conf[param]}", log_file)
    except KeyError:
        ut.printlog(f"{param}: NONE", log_file)

################################################################################
# Begin script

# ***************************************************
#    If photometry path does not exist -> create
# ***************************************************


def create_master_path():

    print("")
    ut.printlog(('********** '
                 'Generating Master Photometry paths '
                 '**********'),
                 log_file)
    print("")

    # Create single mode master photometry path
    if 'photometry_single' in conf['run_steps']:
        ut.makedir(single_master_path)

    # Create dual mode master photometry path
    if 'photometry_dual' in conf['run_steps']:
        ut.makedir(dual_master_path)

    # Create psf mode master photometry path
    if 'photometry_psf' in conf['run_steps']:
        ut.makedir(psf_master_path)


create_master_path()

# ***************************************************
#    Prepare photometry-only dual catalog
# ***************************************************


def extract_photometry_only_dual_catalog():
    """
    Extracts only ID, RA, DEC and PStotal photometry from the dual mode
    catalogs
    """

    print("")
    ut.printlog(('********** '
                 'Creating photometry only dual mode catalogs '
                 '**********'),
                log_file)
    print("")

    for filt in conf['filters']:
        # Generate location of output catalog file
        catalog_name = f"sex_{field}_{filt}_dual_apercorr.fits"
        catalog_file = os.path.join(dual_catalog_path, catalog_name)

        save_file_name = f"{field}_{filt}_photometry_only_dual.fits"
        save_file = os.path.join(dual_master_path, save_file_name)

        if not os.path.exists(save_file):

            ut.extract_sex_photometry(catalog   = catalog_file,
                                      save_file = save_file,
                                      filt      = filt)

            ut.printlog(f"Created file {save_file}", log_file)

        else:
            ut.printlog(f"Catalog {save_file_name} already exists.", log_file)


if 'photometry_dual' in conf['run_steps']:
    extract_photometry_only_dual_catalog()


# ***************************************************
#    Prepare master catalog dual mode
# ***************************************************

def create_master_photometry_dual():
    """
    Combines the individual photometry-only dual mode catalogs into a
    single file
    """

    print("")
    ut.printlog(('********** '
                 'Creating photometry-only dual mode master catalog '
                 '**********'),
                log_file)
    print("")

    save_file_name = f"{field}_master_photometry_only_dual_raw.fits"
    save_file = os.path.join(dual_master_path, save_file_name)

    if not os.path.exists(save_file):

        cmd = f"java -jar {conf['path_to_stilts']} tmatchn "
        cmd += f"nin={len(conf['filters'])} "

        for i in range(len(conf['filters'])):
            filt = conf['filters'][i]

            catalog_name = f"{field}_{filt}_photometry_only_dual.fits"
            catalog_file = os.path.join(dual_master_path, catalog_name)

            cmd += f"ifmt{i+1}=fits in{i+1}={catalog_file} "
            cmd += f"values{i+1}='RA DEC' join{i+1}=always "

        cmd += f"out={save_file} ofmt=fits "
        cmd += f"matcher=sky params=1 multimode=group "

        ut.printlog(cmd, log_file)
        os.system(cmd)

    else:
        ut.printlog(f"Master catalog {save_file_name} already exists.",
                    log_file)


if 'photometry_dual' in conf['run_steps']:
    create_master_photometry_dual()


# ***************************************************
#    Format master dual
# ***************************************************

def format_photometry_master_dual():
    print("")
    ut.printlog(('********** '
                 'Formating photometry-only dual master catalog '
                 '**********'),
                log_file)
    print("")

    catalog_name = f"{field}_master_photometry_only_dual_raw.fits"
    catalog_file = os.path.join(dual_master_path, catalog_name)

    save_file_name = f"{field}_master_photometry_only_dual.fits"
    save_file = os.path.join(dual_master_path, save_file_name)

    if not os.path.exists(save_file):

        ut.format_master_photometry(catalog   = catalog_file,
                                    save_file = save_file,
                                    filters   = conf['filters'],
                                    sexmode   = 'dual',
                                    field     = field,
                                    drname    = conf['data_release_name'])

        ut.printlog(f"Created catalog {save_file}", log_file)

    else:
        ut.printlog(f"Master catalog {save_file_name} already exists.",
                    log_file)


if 'photometry_dual' in conf['run_steps']:
    format_photometry_master_dual()


# ***************************************************
#    Prepare photometry-only single catalog
# ***************************************************

def extract_photometry_only_single_catalog():
    """
    Extracts only ID, RA, DEC and PStotal photometry from the single mode
    catalogs
    """

    print("")
    ut.printlog(('********** '
                 'Creating photometry only single mode catalogs '
                 '**********'),
                log_file)
    print("")

    for filt in conf['filters']:
        # Generate location of output catalog file
        catalog_name = f"sex_{field}_{filt}_single_apercorr.fits"
        catalog_file = os.path.join(single_catalog_path, catalog_name)

        save_file_name = f"{field}_{filt}_photometry_only_single.fits"
        save_file = os.path.join(single_master_path, save_file_name)

        if not os.path.exists(save_file):

            ut.extract_sex_photometry(catalog   = catalog_file,
                                      save_file = save_file,
                                      filt      = filt)

            ut.printlog(f"Created file {save_file}", log_file)

        else:
            ut.printlog(f"Catalog {save_file_name} already exists.", log_file)


if 'photometry_single' in conf['run_steps']:
    extract_photometry_only_single_catalog()


# ***************************************************
#    Prepare master catalog single mode
# ***************************************************

def create_master_photometry_single():
    """
    Combines the individual photometry-only single mode catalogs into a
    single file
    """

    print("")
    ut.printlog(('********** '
                 'Creating photometry-only single mode master catalog '
                 '**********'),
                log_file)
    print("")

    save_file_name = f"{field}_master_photometry_only_single_raw.fits"
    save_file = os.path.join(single_master_path, save_file_name)

    if not os.path.exists(save_file):

        cmd = f"java -jar {conf['path_to_stilts']} tmatchn "
        cmd += f"nin={len(conf['filters'])} "

        for i in range(len(conf['filters'])):
            filt = conf['filters'][i]

            catalog_name = f"{field}_{filt}_photometry_only_single.fits"
            catalog_file = os.path.join(single_master_path, catalog_name)

            cmd += f"ifmt{i+1}=fits in{i+1}={catalog_file} "
            cmd += f"values{i+1}='RA DEC' join{i+1}=always "

        cmd += f"out={save_file} ofmt=fits "
        cmd += f"matcher=sky params=1 multimode=group "

        ut.printlog(cmd, log_file)
        os.system(cmd)

    else:
        ut.printlog(f"Master catalog {save_file_name} already exists.",
                    log_file)


if 'photometry_single' in conf['run_steps']:
    create_master_photometry_single()


# ***************************************************
#    Format master single
# ***************************************************

def format_photometry_master_single():
    print("")
    ut.printlog(('********** '
                 'Formating photometry-only single master catalog '
                 '**********'),
                log_file)
    print("")

    catalog_name = f"{field}_master_photometry_only_single_raw.fits"
    catalog_file = os.path.join(single_master_path, catalog_name)

    save_file_name = f"{field}_master_photometry_only_single.fits"
    save_file = os.path.join(single_master_path, save_file_name)

    if not os.path.exists(save_file):

        ut.format_master_photometry(catalog=catalog_file,
                                    save_file=save_file,
                                    filters=conf['filters'],
                                    sexmode   = 'single',
                                    field     = field,
                                    drname    = conf['data_release_name'])

        ut.printlog(f"Created catalog {save_file}", log_file)

    else:
        ut.printlog(f"Master catalog {save_file_name} already exists.",
                    log_file)


if 'photometry_single' in conf['run_steps']:
    format_photometry_master_single()


# ***************************************************
#    Prepare photometry-only psf catalog
# ***************************************************

def extract_photometry_only_psf_catalog():
    """
    Extracts only ID, RA, DEC and PStotal photometry from the single mode
    catalogs
    """

    print("")
    ut.printlog(('********** '
                 'Creating photometry only psf catalogs '
                 '**********'),
                log_file)
    print("")

    for filt in conf['filters']:
        # Generate location of output catalog file
        catalog_name = f"{field}_{filt}_psf.cat"
        catalog_file = os.path.join(psf_catalog_path, catalog_name)

        save_file_name = f"{field}_{filt}_photometry_only_psf.fits"
        save_file = os.path.join(psf_master_path, save_file_name)

        if not os.path.exists(save_file):

            ut.extract_psf_photometry(catalog   = catalog_file,
                                      save_file = save_file,
                                      filt      = filt)

            ut.printlog(f"Created file {save_file}", log_file)

        else:
            ut.printlog(f"Catalog {save_file_name} already exists.", log_file)


if 'photometry_psf' in conf['run_steps']:
    extract_photometry_only_psf_catalog()


# ***************************************************
#    Prepare master catalog psf
# ***************************************************

def create_master_photometry_psf():
    """
    Combines the individual photometry-only psf catalogs into a single file
    """

    print("")
    ut.printlog(('********** '
                 'Creating photometry-only psf master catalog '
                 '**********'),
                log_file)
    print("")

    save_file_name = f"{field}_master_photometry_only_psf_raw.fits"
    save_file = os.path.join(psf_master_path, save_file_name)

    if not os.path.exists(save_file):

        cmd = f"java -jar {conf['path_to_stilts']} tmatchn "
        cmd += f"nin={len(conf['filters'])} "

        for i in range(len(conf['filters'])):
            filt = conf['filters'][i]

            catalog_name = f"{field}_{filt}_photometry_only_psf.fits"
            catalog_file = os.path.join(psf_master_path, catalog_name)

            cmd += f"ifmt{i+1}=fits in{i+1}={catalog_file} "
            cmd += f"values{i+1}='RA DEC' join{i+1}=always "

        cmd += f"out={save_file} ofmt=fits "
        cmd += f"matcher=sky params=1 multimode=group "

        ut.printlog(cmd, log_file)
        os.system(cmd)

    else:
        ut.printlog(f"Master catalog {save_file_name} already exists.",
                    log_file)


if 'photometry_psf' in conf['run_steps']:
    create_master_photometry_psf()


# ***************************************************
#    Format master psf
# ***************************************************

def format_photometry_master_psf():
    print("")
    ut.printlog(('********** '
                 'Formating photometry-only psf master catalog '
                 '**********'),
                log_file)
    print("")

    catalog_name = f"{field}_master_photometry_only_psf_raw.fits"
    catalog_file = os.path.join(psf_master_path, catalog_name)

    save_file_name = f"{field}_master_photometry_only_psf.fits"
    save_file = os.path.join(psf_master_path, save_file_name)

    if not os.path.exists(save_file):

        ut.format_master_photometry(catalog=catalog_file,
                                    save_file=save_file,
                                    filters=conf['filters'],
                                    sexmode   = 'psf',
                                    field     = field,
                                    drname    = conf['data_release_name'])

        ut.printlog(f"Created catalog {save_file}", log_file)

    else:
        ut.printlog(f"Master catalog {save_file_name} already exists.",
                    log_file)


if 'photometry_psf' in conf['run_steps']:
    format_photometry_master_psf()



# -*- coding: utf-8 -*-

# ******************************************************************************
#                          S-PLUS CALIBRATION PIPELINE
#                               quality_check.py
#            Produces the quality check plot in the end of the pipeline
# ******************************************************************************

"""
Produces the quality check plot in the end of the pipeline

The S-PLUS field is given as the first command line argument. Configurations
are set in the config file, given as the second command line argument.

--------------------------------------------------------------------------------
   FUNCTIONS:
--------------------------------------------------------------------------------


--------------------------------------------------------------------------------
   COMMENTS:
--------------------------------------------------------------------------------
Ideally this script should only be run through the pipeline.py script.

Assumes that all calibration steps have been run for this field

--------------------------------------------------------------------------------
   USAGE:
--------------------------------------------------------------------------------
$python3 quality_check.py *field_name* *config_file*

----------------
"""

################################################################################
# Import external packages

import os
import sys
import numpy as np
from matplotlib import pyplot as plt
from astropy.table import Table

pipeline_path = os.path.split(__file__)[0]
spluscalib_path = os.path.split(pipeline_path)[0]

sys.path.append(spluscalib_path)

################################################################################
# Import spluscalib packages

from spluscalib import utils as ut

################################################################################
# Read parameters

field     = sys.argv[1]
conf_file = sys.argv[2]

conf = ut.pipeline_conf(conf_file)

try:
    quality_path = sys.argv[3]
except IndexError:
    quality_path = conf['quality_path'] 


################################################################################
# Get directories
suffix = ut.calibration_suffix(conf)

field_path       = os.path.join(conf['run_path'], field)
photometry_path  = os.path.join(field_path, 'Photometry')
crossmatch_path  = os.path.join(field_path, 'Crossmatch')
calibration_path = os.path.join(field_path, f'Calibration_{suffix}')

################################################################################
# Check if plot already exists

save_file = os.path.join(calibration_path, f'{field}_plots_PSF.png')

if os.path.exists(save_file):
    # Copy to quality_path
    if quality_path:
        
        suffix = ut.calibration_suffix(conf)
        extinction = conf['extinction_correction']
        save_file_quality_path = os.path.join(quality_path,
                                              f'{field}_{suffix}_{extinction}_PSF.png')
        
        cmd = f"cp {save_file} {save_file_quality_path}"
        os.system(cmd)

    msg = "quality plot already exists and will not be made again"
    raise FileExistsError(msg)

################################################################################
# Initiallize plots

nrow = 2
ncol = 3

fig, axs = plt.subplots(nrow, ncol, figsize=[15,10.5])
for ax in fig.axes:
    ax.tick_params(labelrotation=90, axis = 'y')

fig.suptitle(field, fontsize=14)

plt.subplots_adjust(top=0.95)

filt_colors = {'U': "#d600a6",
               'F378': "#c300bc",
               'F395': "#a700ce",
               'F410': "#8300de",
               'F430': "#3e00f3",
               'G': "#00c1ff",
               'F515': "#17ff00",
               'R': "#ff5f00",
               'F660': "#f80000",
               'I': "#cb0000",
               'F861': "#a30000",
               'Z': "#9b0000"}


################################################################################
# Plot psf mode detections

def plot_psf_mode_detections():

    psf_master = os.path.join(photometry_path, 'psf', 'master',
                                 f"{field}_master_photometry_only_psf.fits")

    # Load catalog
    psf_master_data = ut.load_data(psf_master)

    x = psf_master_data.loc[:,'X'].values
    y = psf_master_data.loc[:,'Y'].values

    axs[0,0].scatter(x,y, s = 0.5, alpha = 0.1)
    axs[0,0].set_xlabel("X")
    axs[0,0].set_ylabel("Y")
    axs[0,0].set_xlim([0, 11000])
    axs[0,0].set_ylim([0, 11000])

    axs[0,0].text(100, 10500, 'Photometry master [psf]')
    axs[0,0].text(100, 100, f'N_sources: {len(psf_master_data)}')
try:
    plot_psf_mode_detections()
except:
    pass


################################################################################
# Plot crossmatch

def plot_crossmatch():

    correction = conf['extinction_correction']

    suffix = ut.calibration_suffix(conf)

    crossmatch_name = ut.crossmatch_catalog_name(field, conf)
    crossmatch_name = crossmatch_name.replace(".fits",
                                              f"_ebvcorr_{correction}.fits")

    crossmatch = os.path.join(crossmatch_path, crossmatch_name)

    # Load catalog
    crossmatch_data = ut.load_data(crossmatch)

    x = crossmatch_data.loc[:,'X'].values
    y = crossmatch_data.loc[:,'Y'].values

    axs[0,1].scatter(x,y, s = 2, alpha = 1)
    axs[0,1].set_xlabel("X")
    axs[0,1].set_ylabel("Y")
    axs[0,1].set_xlim([0, 11000])
    axs[0,1].set_ylim([0, 11000])

    txt = (f"Crossmatch [ {conf['calibration_photometry']} | "
           f"{' '.join(conf['reference_catalog'])} | "
           f"{conf['extinction_correction']} ]")
    axs[0,1].text(100, 10500, txt)
    axs[0,1].text(100, 100, f'N_sources: {len(crossmatch_data)}')

try:
    plot_crossmatch()
except:
    pass


################################################################################
# Plot internal SED fitting


def plot_internal_sed_fitting():

    filters = conf['filters']
    shift = [6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5]
    mag_cut = conf['zp_fitting_mag_cut']

    sed_fit_file = os.path.join(calibration_path, 'internal',
                                f"{field}_mag_int.cat")

    cat_data = ut.load_data(sed_fit_file)

    ax = axs[0,2]
    ax.set_xlim([11,21])
    ax.set_ylim([-6,7])

    ax.set_xlabel("model mag")
    ax.set_ylabel("model mag - splus mag")

    ax.set_yticks([-5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6])
    ax.set_yticklabels(['', '', '', '', -0.5, 0, 0.5, '', '', '', '', ''],
                       minor=False)

    for i in range(len(filters)):
        filt = filters[i]
        color = filt_colors[filt]

        # Remove 99
        data_selection = np.abs(cat_data.loc[:, f'SPLUS_{filt}'] < 50)
        cat_data = cat_data[data_selection]

        ax.text(12.3, shift[i]+0.1, filt, color = color)

        ##############
        # Prepare data

        x = cat_data.loc[:, f'SPLUS_{filt}_mod']
        y = 2*(x - (cat_data.loc[:, f'SPLUS_{filt}']))

        dwarfs = cat_data.loc[:, 'logg'].values > 3
        selection = (x >= mag_cut[0]) & (x <= mag_cut[1]) & dwarfs

        ############################
        # Plot

        ax.scatter(x[selection], y[selection] + shift[i],
                   c=color, s=5, alpha=0.5, zorder=2)

        ax.scatter(x[~selection], y[~selection] + shift[i],
                   c="#666666", s=2, alpha=0.1, zorder=1)

        ax.plot([8,22],[shift[i],shift[i]], '--', color = 'black', alpha = 0.5,
                zorder = 0, linewidth = 0.5)

        ax.plot([mag_cut[0], mag_cut[0]], [-6,7],
                 color='#0066FF', linestyle='-', linewidth=0.5, zorder=6)

        ax.plot([mag_cut[1], mag_cut[1]], [-6,7],
                 color='#0066FF', linestyle='-', linewidth=0.5, zorder=6)


try:
    plot_internal_sed_fitting()
except:
    pass


################################################################################
# Plot color-color 1

def plot_color_color_1():

    ax = axs[1,0]
    ax.set_xlim([0.75,3])
    ax.set_ylim([0.5,4])

    ax.set_xlabel("U - G")
    ax.set_ylabel("F378 - F660")

    stellar_locus_reference_file = conf['stellar_locus_reference']
    sl_ref = ut.load_data(stellar_locus_reference_file)

    # Plot reference stellar locus
    x_ref = sl_ref.loc[:,'SPLUS_U'].values - sl_ref.loc[:,'SPLUS_G'].values
    y_ref = sl_ref.loc[:,'SPLUS_F378'].values - sl_ref.loc[:,'SPLUS_F660'].values

    ax.scatter(x_ref, y_ref, color = "#AAAAAA", s = 2, alpha = 0.05, zorder = -1)

    # Load field data (calibrated and EB-V corrected)
    catalog_file = os.path.join(calibration_path, 'internal',
                                f'{field}_mag_int.cat')

    cat_data = ut.load_data(catalog_file)

    x1 = cat_data.loc[:,'SPLUS_U'].values
    x2 = cat_data.loc[:,'SPLUS_G'].values
    y1 = cat_data.loc[:,'SPLUS_F378'].values
    y2 = cat_data.loc[:,'SPLUS_F660'].values

    f = (x1 < 19) & (x2 < 19) & (y1 < 19) & (y2 < 19)

    x_cat = x1[f] - x2[f]
    y_cat = y1[f] - y2[f]

    ax.scatter(x_cat, y_cat, color = "#FF5F00", s = 5, alpha = 0.2, zorder = 1)

try:
    plot_color_color_1()
except:
    pass

################################################################################
# Plot color-color 2

def plot_color_color_2():

    ax = axs[1,1]
    ax.set_xlim([0.5,3.5])
    ax.set_ylim([0,3.5])

    ax.set_xlabel("F395 - R")
    ax.set_ylabel("F410 - F861")

    stellar_locus_reference_file = conf['stellar_locus_reference']
    sl_ref = ut.load_data(stellar_locus_reference_file)

    # Plot reference stellar locus
    x_ref = sl_ref.loc[:,'SPLUS_F395'].values - sl_ref.loc[:,'SPLUS_R'].values
    y_ref = sl_ref.loc[:,'SPLUS_F410'].values - sl_ref.loc[:,'SPLUS_F861'].values

    ax.scatter(x_ref, y_ref, color = "#AAAAAA", s = 2, alpha = 0.05, zorder = -1)

    # Load field data (calibrated and EB-V corrected)
    catalog_file = os.path.join(calibration_path, 'internal',
                                f'{field}_mag_int.cat')

    cat_data = ut.load_data(catalog_file)

    x1 = cat_data.loc[:,'SPLUS_F395'].values
    x2 = cat_data.loc[:,'SPLUS_R'].values
    y1 = cat_data.loc[:,'SPLUS_F410'].values
    y2 = cat_data.loc[:,'SPLUS_F861'].values

    f = (x1 < 19) & (x2 < 19) & (y1 < 19) & (y2 < 19)

    x_cat = x1[f] - x2[f]
    y_cat = y1[f] - y2[f]

    ax.scatter(x_cat, y_cat, color = "#FF5F00", s = 5, alpha = 0.2, zorder = 1)

try:
    plot_color_color_2()
except:
    pass


################################################################################
# Plot color-color 3

def plot_color_color_3():

    ax = axs[1,2]
    ax.set_xlim([0.3,2.5])
    ax.set_ylim([0,2])

    ax.set_xlabel("F430 - I")
    ax.set_ylabel("F515 - Z")

    stellar_locus_reference_file = conf['stellar_locus_reference']
    sl_ref = ut.load_data(stellar_locus_reference_file)

    # Plot reference stellar locus
    x_ref = sl_ref.loc[:,'SPLUS_F430'].values - sl_ref.loc[:,'SPLUS_I'].values
    y_ref = sl_ref.loc[:,'SPLUS_F515'].values - sl_ref.loc[:,'SPLUS_Z'].values

    ax.scatter(x_ref, y_ref, color = "#AAAAAA", s = 2, alpha = 0.05, zorder = -1)

    # Load field data (calibrated and EB-V corrected)
    catalog_file = os.path.join(calibration_path, 'internal',
                                f'{field}_mag_int.cat')

    cat_data = ut.load_data(catalog_file)

    x1 = cat_data.loc[:,'SPLUS_F430'].values
    x2 = cat_data.loc[:,'SPLUS_I'].values
    y1 = cat_data.loc[:,'SPLUS_F515'].values
    y2 = cat_data.loc[:,'SPLUS_Z'].values

    f = (x1 < 19) & (x2 < 19) & (y1 < 19) & (y2 < 19)

    x_cat = x1[f] - x2[f]
    y_cat = y1[f] - y2[f]

    ax.scatter(x_cat, y_cat, color = "#FF5F00", s = 5, alpha = 0.2, zorder = 1)

try:
    plot_color_color_3()
except:
    pass

plt.tight_layout()

plt.savefig(save_file)

if quality_path:
    suffix = ut.calibration_suffix(conf)
    extinction = conf['extinction_correction']
    save_file_quality_path = os.path.join(quality_path,
                                         f'{field}_{suffix}_{extinction}.png')

    plt.savefig(save_file_quality_path)


plt.clf()
plt.close()



# -*- coding: utf-8 -*-

# ******************************************************************************
#                          S-PLUS CALIBRATION PIPELINE
#                             photometry_master.py
#                      Generate photometry master catalogs
# ******************************************************************************

"""
Crossmatches S-PLUS with the reference catalog

The S-PLUS field is given as the first command line argument. Configurations
are set in the config file, given as the second command line argument.

--------------------------------------------------------------------------------
   FUNCTIONS:
--------------------------------------------------------------------------------

copy_splus_master_photometry()
dowload_references()
crossmatch_all_references()

--------------------------------------------------------------------------------
   COMMENTS:
--------------------------------------------------------------------------------
Ideally this script should only be run through the pipeline.py script.

Assumes that photometry_master.py has already been run for this field.

--------------------------------------------------------------------------------
   USAGE:
--------------------------------------------------------------------------------
$python3 crossmatch.py *field_name* *config_file*

----------------
"""

################################################################################
# Import external packages

import os
import sys

steps_path = os.path.split(__file__)[0]
pipeline_path = os.path.split(steps_path)[0]
spluscalib_path = os.path.split(pipeline_path)[0]

sys.path.append(spluscalib_path)

################################################################################
# Import spluscalib packages

from spluscalib import utils as ut

################################################################################
# Read parameters

field     = sys.argv[1]
conf_file = sys.argv[2]

conf = ut.pipeline_conf(conf_file)

################################################################################
# Get directories

field_path = os.path.join(conf['run_path'], field)

suffix = ut.calibration_suffix(conf)
calibration_path = os.path.join(field_path, f'Calibration_{suffix}')
crossmatch_path  = os.path.join(field_path, 'Crossmatch')

photometry_path = os.path.join(field_path, 'Photometry')
images_path     = os.path.join(field_path, 'Images')

log_path = os.path.join(crossmatch_path, 'logs')


################################################################################
# Initiate log file

ut.makedir(crossmatch_path)
ut.makedir(log_path)

log_file_name = os.path.join(log_path, 'crossmatch.log')
log_file_name = ut.gen_logfile_name(log_file_name)
log_file = os.path.join(crossmatch_path, log_file_name)

with open(log_file, "w") as log:
    log.write("")


################################################################################
# Log configuration

ut.printlog("Reference crossmatch parameters:", log_file)

crossmatch_params = ['run_path', 'filters', 'calibration_photometry',
                     'reference_catalog']

for param in crossmatch_params:
    try:
        ut.printlog(f"{param}: {conf[param]}", log_file)
    except KeyError:
        ut.printlog(f"{param}: NONE", log_file)

################################################################################
# Begin script

# ***************************************************
#    Copy splus master photometry catalog
# ***************************************************


def copy_splus_master_photometry():

    """
    Copy S-PLUS photometry to the crossmatch path
    """

    print("")
    ut.printlog(('********** '
                 'Copying S-PLUS master photometry '
                 '**********'),
                 log_file)
    print("")

    calib_phot = conf['calibration_photometry']

    catalog_name = f"{field}_master_photometry_only_{calib_phot}.fits"
    catalog_file = os.path.join(photometry_path, calib_phot, 'master',
                                catalog_name)

    save_name = f'{field}_SPLUS_{calib_phot}_tmp.fits'
    save_file = os.path.join(crossmatch_path, save_name)

    if not os.path.exists(save_file):

        cmd = f"cp {catalog_file} {save_file}"
        ut.printlog(cmd, log_file)
        os.system(cmd)

    else:
        ut.printlog(f"Catalog {save_file} already exists", log_file)


copy_splus_master_photometry()


# ***************************************************
#    Add photometry offset
# ***************************************************

def adding_instrumental_offset():

    calib_phot = conf['calibration_photometry']

    catalog_name = f'{field}_SPLUS_{calib_phot}_tmp.fits'
    catalog_file = os.path.join(crossmatch_path, catalog_name)

    save_name = f'{field}_SPLUS_{calib_phot}.fits'
    save_file = os.path.join(crossmatch_path, save_name)

    if not os.path.exists(save_file):
        if conf["offset_inst_mag"].lower() != "none":

            print("")
            ut.printlog(('********** '
                         'Adding instrumental photometry offset '
                         '**********'),
                        log_file)
            print("")

            ut.add_inst_offset(catalog     = catalog_file,
                               offset_file = conf["offset_inst_mag"],
                               save_file   = save_file)

        else:
            cmd = f"cp {catalog_file} {save_file}"
            ut.printlog(cmd, log_file)
            os.system(cmd)

    else:
        ut.printlog(f"Catalog {save_file} already exists", log_file)


adding_instrumental_offset()

# ***************************************************
#    Download reference catalogs
# ***************************************************


def dowload_references():

    """
    Downloads reference catalogs covering the field's region
    """

    print("")
    ut.printlog(('********** '
                 'Downloading reference catalogs '
                 '**********'),
                 log_file)
    print("")

    for ref in conf['reference_catalog']:

        ut.printlog(f"Trying to download {ref}", log_file)

        ref_catalog = f'{field}_{ref}.fits'
        save_file = os.path.join(crossmatch_path, ref_catalog)

        # S-PLUS image to extract header info
        filt = conf['filters'][-1]
        image_name = f'{field}_{filt}_swp.fz'
        image_file = os.path.join(images_path, image_name)

        #fieldID_name =  f'{field}_FIELD_ID.fits'
        #fieldID_file = os.path.join(photometry_path, fieldID_name)

        if not os.path.exists(save_file):

            ut.download_reference(image = image_file,
                                  reference = ref,
                                  save_file = save_file)

            ut.printlog(f"Downloaded {ref} catalog for field {field}", log_file)

        else:
            ut.printlog(f"Reference {ref} already downloaded for field {field}",
                        log_file)

        print("")


dowload_references()


# ***************************************************
#    Download GAIA DR2 for the field
# ***************************************************

def dowload_gaiadr2():

    """
    Downloads GAIA DR2 catalogs covering the field's region
    """

    print("")
    ut.printlog(('********** '
                 'Downloading GAIA DR2 catalog '
                 '**********'),
                 log_file)
    print("")

    ref_catalog = f'{field}_GAIADR2_VEGA.fits'
    save_file = os.path.join(crossmatch_path, ref_catalog)

    # S-PLUS image to extract header info
    filt = conf['filters'][-1]
    image_name = f'{field}_{filt}_swp.fz'
    image_file = os.path.join(images_path, image_name)

    #fieldID_name = f'{field}_FIELD_ID.fits'
    #fieldID_file = os.path.join(photometry_path, fieldID_name)

    if not os.path.exists(save_file):

        ut.download_reference(image = image_file,
                              reference = 'GAIADR2',
                              save_file = save_file)

        ut.printlog(f"Downloaded Gaia DR2 catalog for field {field}", log_file)

    else:
        ut.printlog(f"Gaia DR2 already downloaded for field {field}",
                    log_file)

    print("")


dowload_gaiadr2()


# ***************************************************
#    Convert GAIA to AB magnitudes
# ***************************************************

def convert_gaia_VEGA_to_AB():

    """
    Converts Gaia magnitudes to AB system
    """

    print("")
    ut.printlog(('********** '
                 'Converting Gaia magnitudes to AB system '
                 '**********'),
                 log_file)
    print("")

    catalog_name = f'{field}_GAIADR2_VEGA.fits'
    catalog_file = os.path.join(crossmatch_path, catalog_name)

    save_name = f'{field}_GAIADR2.fits'
    save_file = os.path.join(crossmatch_path, save_name)

    if not os.path.exists(save_file):

        ut.convert_gaia_Vega2AB(gaia_catalog = catalog_file,
                                save_file    = save_file)

        ut.printlog(f"Converted Gaia DR2 magnitudes to AB for field {field}",
                    log_file)

    else:
        ut.printlog(f"Gaia DR2 already converted to AB for field {field}",
                    log_file)

    print("")


convert_gaia_VEGA_to_AB()

# ***************************************************
#    Crossmatch between references if more than one
# ***************************************************

def crossmatch_all_references():

    """
    Combines S-PLUS and reference catalogs
    """

    print("")
    ut.printlog(('********** '
                 'Crossmatching all reference catalogs '
                 '**********'),
                 log_file)
    print("")

    # Get photometry mode used for calibration
    calib_phot = conf['calibration_photometry']

    # Get number of reference catalogs
    nref = len(conf['reference_catalog'])

    # Include S-PLUS catalog in the tmatchn
    splus_cat = f'{field}_SPLUS_{calib_phot}.fits'
    splus_cat_file = os.path.join(crossmatch_path, splus_cat)

    # Include GAIA DR2
    gaia_cat = f'{field}_GAIADR2.fits'
    gaia_cat_file = os.path.join(crossmatch_path, gaia_cat)

    # Start save name
    save_name = ut.crossmatch_catalog_name(field, conf)
    save_file = os.path.join(crossmatch_path, save_name)

    if not os.path.exists(save_file):

        # Start tmatchn cmd
        cmd = f"java -jar {conf['path_to_stilts']} tmatchn "
        cmd += f"nin={nref + 2} "

        cmd += f"ifmt1=fits in1={splus_cat_file} "
        cmd += f"values1='RAJ2000 DEJ2000' join1=match "

        cmd += f"ifmt2=fits in2={gaia_cat_file} "
        cmd += f"values2='GAIADR2_RAJ2000 GAIADR2_DEJ2000' join2=match "

        for i in range(nref):
            ref = conf['reference_catalog'][i]

            ref_cat = f'{field}_{ref}.fits'
            ref_cat_file = os.path.join(crossmatch_path, ref_cat)

            values = f"{ref}_RAJ2000 {ref}_DEJ2000"
            cmd += f"ifmt{i + 3}=fits in{i + 3}={ref_cat_file} "
            cmd += f"values{i + 3}='{values}' join{i + 3}=match "

        cmd += f"out={save_file} ofmt=fits "
        cmd += f"matcher=sky params=3 multimode=group "

        ut.printlog(cmd, log_file)
        os.system(cmd)

        ut.printlog(f"Created catalog {save_file}", log_file)

    else:
        ut.printlog(f"Catalog {save_name} already exists", log_file)


crossmatch_all_references()


# ***************************************************
#    Correct extinction
# ***************************************************

def correct_extinction_crossmatched_catalog():
    """
    Generates crossmatched catalog with extinction correction
    """

    print("")
    ut.printlog(('********** '
                 'Applying exctinction correction '
                 '**********'),
                 log_file)
    print("")

    ###############################
    # Get crossmatched catalog file
    cmatch_name = ut.crossmatch_catalog_name(field, conf)
    cmatch_file = os.path.join(crossmatch_path, cmatch_name)

    ##################
    # Apply correction
    correction = conf['extinction_correction']
    ebv_map_path = conf['extinction_maps_path']

    save_file = cmatch_file.replace(".fits", f"_ebvcorr_{correction}.fits")

    if not os.path.exists(save_file):
        ut.correct_extinction(catalog = cmatch_file,
                              save_file = save_file,
                              correction = correction,
                              ebv_maps_path = ebv_map_path)

        ut.printlog(f"Created file {save_file}", log_file)

    else:
        ut.printlog(f"Extinction corrected catalog already exists", log_file)


if conf['extinction_correction'].lower() != 'none':
    correct_extinction_crossmatched_catalog()



# -*- coding: utf-8 -*-

# ******************************************************************************
#                          S-PLUS CALIBRATION PIPELINE
#                                 pipeline.py
#             Runs all the steps of the S-PLUS calibration pipeline
# ******************************************************************************


"""
This scripts runs the steps of the calibration pipeline

--------------------------------------------------------------------------------
   INPUTS:
--------------------------------------------------------------------------------
All inputs are taken from the config_file

--------------------------------------------------------------------------------
   COMMENTS:
--------------------------------------------------------------------------------
User must not change anything in this code. Changes to choose which steps to
run should be made in the config_file

config_file.py
    run_steps        = [*list of stepś*]

    possible steps are:
        The following steps are necessary to complete the calibration
        'photometry_single': obtain aperture photometry in single mode
        'photometry_dual'  : obtain aperture photometry in dual mode
        'photometry_psf'   : obtain psf photometry
        'correction_xy'    : corrects XY position zero-points inhomogeneities
        'aper_corretion'   : applies aperture correction to fixed aperture

--------------------------------------------------------------------------------
   USAGE:
--------------------------------------------------------------------------------
$ python pipeline_final *field_list_file* *config_file*

the field_list_file should be an ascii file where each line is the name of one
field to calibrate.
----------------
"""

################################################################################
# Import external packages

import os
import sys

from termcolor import cprint

usage_message = """
USAGE:

option1: list of fields
$spluscalib  *field_list_file* *config_file*

option2: single field
$spluscalib  *field_name* *config_file* --onefield

"""

pipeline_path = os.path.split(__file__)[0]

sys.path.append(pipeline_path)

################################################################################
# Import spluscalib packages

import utils as ut

################################################################################
# Read parameters

cprint("""
************************************************************************
*                                                                      *
*                   Starting the calibration pipeline                  *
*                                                                      *
************************************************************************
""", color = "cyan")

cwd = os.getcwd()

try:
    field_list_file = os.path.join(cwd, sys.argv[1])
    conf_file = os.path.join(cwd, sys.argv[2])

except IndexError:
    print(usage_message)
    sys.exit()

pipeline_path = os.path.split(__file__)[0]

cprint(conf_file, color = "white", attrs=["dark"])
cprint("******* reading the configuration file *******", color = "yellow")
conf = ut.pipeline_conf(conf_file)

################################################################################
# Create save path directory

ut.makedir(conf['save_path'])
ut.makedir(conf['run_path'])

logs_path = os.path.join(conf['run_path'], 'logs')
ut.makedir(logs_path)

################################################################################
# Initiate log file

log_file_name = os.path.join(logs_path, 'pipeline.log')
log_file_name = ut.gen_logfile_name(log_file_name)
log_file = os.path.join(conf['run_path'], log_file_name)

with open(log_file, "w") as log:
    log.write("")

################################################################################
# Reading parameters

print("\n\n")

ut.printlog("*********** Reading configuratrion file **********", log_file, 
            color = "yellow")

for param in list(conf.keys()):
    ut.printlog(f"{param}: {conf[param]}", log_file, 
                color = "blue")

print("\n\n")

################################################################################
# Reading field list

ut.printlog("*********** Reading field list **********", log_file,
           color = "yellow")

if '--onefield' in sys.argv:
    fields = [sys.argv[1]]
else:
    fields = ut.load_field_list(field_list_file)

ut.printlog(f"Running the pipeline for fields:", log_file, color = "yellow")
ut.printlog(f"{fields}", log_file, 
            color = "yellow", attrs = ["bold"])

################################################################################
# Running pipeline for each field

for field in fields:

    ut.printlog(f"""
    ************************************************************************
    *                                                                      *
    *                Working on field {field:37}*
    *                                                                      *
    ************************************************************************
    """, log_file, color = "cyan", attrs = ["bold"])

    #################
    # Make field path
    field_path = os.path.join(conf['run_path'], field)
    logs_path = os.path.join(field_path, 'logs')

    field_images_path = os.path.join(field_path, 'Images')
    
    if "--skip_missing_images" in sys.argv:
        source_images_path = os.path.join(conf["path_to_images"], field)
        if not os.path.exists(source_images_path):
            ut.printlog(f"{field} images not found. Skipping to next field\n\n",
                        log_file, color = "white", attrs = ["dark"])
            continue

    if not os.path.exists(field_path):
        ut.makedir(field_path)
        ut.printlog(f"Created directory {field_path}.", log_file,
                    color = "green", attrs=["bold"])

        ut.makedir(logs_path)
        ut.printlog(f"Created directory {logs_path}.", log_file,
                    color = "green", attrs=["bold"])

        ut.makedir(field_images_path)
        ut.printlog(f"Created directory {field_images_path}.", log_file,
                    color = "green", attrs=["bold"])

    else:
        ut.printlog(f"Directory {field_path} already exists.", log_file,
                    color = "white", attrs=["bold", "dark"])

    print("")

    ###################
    # Copy files already present in save_path
    
    if conf['run_path'] != conf['save_path']:
        field_save_path = os.path.join(conf['save_path'], field)
        cmd = f"rsync --update -haz --progress {field_save_path} {conf['run_path']}"
        ut.printlog(cmd, log_file, color = "blue")
        os.system(cmd)
        print("")

    ###################
    # Prepare field log

    log_field_name = os.path.join(logs_path, f'{field}.log')
    log_field_name = ut.gen_logfile_name(log_field_name)
    log_field_file = os.path.join(conf['run_path'], log_field_name)

    with open(log_file, "w") as log:
        log.write("")

    # Make backup of config file
    cmd = f"cp {conf_file} {logs_path}"
    ut.printlog(cmd, log_field_file, color = "green")
    os.system(cmd)

    # Run steps

    ############################################################################
    # Copy fz images and unpack, if necessary

    # Check if configuration asks photometry to be run
    run_photometry = 'photometry_single' in conf['run_steps']
    run_photometry = run_photometry or ('photometry_dual' in conf['run_steps'])
    run_photometry = run_photometry or ('photometry_psf' in conf['run_steps'])

    already_run = False
    already_run_single = False
    already_run_dual = False
    already_run_psf = False

    # Check if photometry was already run
    if run_photometry:

        already_run = True
        if 'photometry_single' in conf['run_steps']:
            already_run_single = ut.check_photometry(field = field,
                                               save_path   = conf['save_path'],
                                               photometry  = 'single',
                                               filter_list = conf['filters'])

            already_run = already_run & already_run_single

        if 'photometry_dual' in conf['run_steps']:
            already_run_dual = ut.check_photometry(field=field,
                                                   save_path=conf['save_path'],
                                                   photometry='dual',
                                                   filter_list=conf['filters'])

            already_run = already_run & already_run_dual

        if 'photometry_psf' in conf['run_steps']:
            already_run_psf = ut.check_photometry(field=field,
                                                  save_path=conf['save_path'],
                                                  photometry='psf',
                                                  filter_list=conf['filters'])

            already_run = already_run & already_run_psf

    if run_photometry and not ('--nofits' in sys.argv) and not already_run:

        for filt in conf['filters']:
            image_name  = f'{field}_{filt}_swp.fz'
            image_db_fz = os.path.join(conf['path_to_images'], field,
                                       image_name)
            image_fz    = os.path.join(field_images_path, image_name)

            ##############################################
            # Copying fz from path_to_images to field path

            if not os.path.exists(image_fz):

                ut.printlog(f"Copying image {image_name}", log_field_file,
                            color = "yellow")

                cmd = f"cp {image_db_fz} {field_images_path}"
                ut.printlog(cmd, log_field_file, color = "green")
                os.system(cmd)

            else:
                ut.printlog(f"Image {image_fz} already exists",
                            log_field_file, color = "white", attrs = ["dark"])
            

            if conf['use_weight']:

                weight_fz = image_fz.replace(".fz", "weight.fz")
                weight_db_fz = image_db_fz.replace(".fz", "weight.fz")

                if not os.path.exists(weight_fz):
                    cmd = f"cp {weight_db_fz} {field_images_path}"
                    ut.printlog(cmd, log_field_file, color = "green")
                    os.system(cmd)

                else:
                    ut.printlog(f"Image {weight_fz} already exists",
                                log_field_file, 
                                color = "white", attrs = ["dark"])

            #######################
            # Unpacking fits images

            image_fits = image_fz.replace(".fz", ".fits")

            if not os.path.exists(image_fits):

                ut.printlog(f"Unpacking image {image_fz}", log_field_file,
                            color = "yellow")
                ut.fz2fits(image_fz)

            else:
                ut.printlog(f"Image {image_fits} already exists",
                            log_field_file, color = "white", attrs = ["dark"])
                
                
            if conf['use_weight']:
                weight_fz = image_fz.replace(".fz", "weight.fz")
                weight_fits = weight_fz.replace(".fz", ".fits")

                if not os.path.exists(weight_fits):             
                    ut.printlog(f"Unpacking weight {weight_fz}",
                                log_field_file, color = "yellow")
                    ut.fz2fits(weight_fz)

                else:
                    ut.printlog(f"Image {weight_fits} already exists",
                                log_field_file, 
                                color = "white", attrs = ["dark"])

            print("")

    ############################################################################

    ############################################################################
    # Run the photometry
    if 'photometry_single' in conf['run_steps'] and not already_run_single:

        final_file = os.path.join(field_path, 'Photometry', 'single', 'master',
                                  f"{field}_master_photometry_only_single.fits")

        if not os.path.exists(final_file):
            script = os.path.join(pipeline_path,'steps','photometry_single.py')
            cmd = f'python3 {script} {field} {conf_file}'
            ut.printlog(cmd, log_field_file, 
                        color = "blue", on_color = "on_white")
            os.system(cmd)

            print("\n\n")

    if 'photometry_dual' in conf['run_steps'] and not already_run_dual:

        final_file = os.path.join(field_path, 'Photometry', 'dual', 'master',
                                  f"{field}_master_photometry_only_dual.fits")

        if not os.path.exists(final_file):
            script = os.path.join(pipeline_path, 'steps', 'photometry_dual.py')
            cmd = f'python3 {script} {field} {conf_file}'
            ut.printlog(cmd, log_field_file,
                        color = "blue", on_color = "on_white")
            os.system(cmd)

            print("\n\n")

    if 'photometry_psf' in conf['run_steps'] and not already_run_psf:

        final_file = os.path.join(field_path, 'Photometry', 'psf', 'master',
                                  f"{field}_master_photometry_only_psf.fits")

        if not os.path.exists(final_file):
            script = os.path.join(pipeline_path, 'steps', 'photometry_psf.py')
            cmd = f'python3 {script} {field} {conf_file}'
            ut.printlog(cmd, log_field_file,
                        color = "blue", on_color = "on_white")
            os.system(cmd)

            print("\n\n")

    if 'correction_xy' in conf['run_steps']:

        script = os.path.join(pipeline_path, 'steps', 'correction_xy.py')
        cmd = f'python3 {script} {field} {conf_file}'
        ut.printlog(cmd, log_field_file,
                    color = "blue", on_color = "on_white")
        os.system(cmd)

        print("\n\n")

    if 'correction_aper' in conf['run_steps']:

        script = os.path.join(pipeline_path, 'steps', 'correction_aper.py')
        cmd = f'python3 {script} {field} {conf_file}'
        ut.printlog(cmd, log_field_file,
                    color = "blue", on_color = "on_white")
        os.system(cmd)

        print("\n\n")

    if 'photometry_id' in conf['run_steps']:

        script = os.path.join(pipeline_path, 'steps', 'photometry_ID.py')
        cmd = f'python3 {script} {field} {conf_file}'
        ut.printlog(cmd, log_field_file,
                    color = "blue", on_color = "on_white")
        os.system(cmd)

        print("\n\n")

    if 'photometry_master' in conf['run_steps']:

        script = os.path.join(pipeline_path, 'steps', 'photometry_master.py')
        cmd = f'python3 {script} {field} {conf_file}'
        ut.printlog(cmd, log_field_file,
                    color = "blue", on_color = "on_white")
        os.system(cmd)

        print("\n\n")

    if 'crossmatch' in conf['run_steps']:

        script = os.path.join(pipeline_path, 'steps', 'crossmatch.py')
        cmd = f'python3 {script} {field} {conf_file}'
        ut.printlog(cmd, log_field_file,
                    color = "blue", on_color = "on_white")
        os.system(cmd)

        print("\n\n")

    if 'calibration_external' in conf['run_steps']:

        script = os.path.join(pipeline_path, 'steps', 'calibration_external.py')
        cmd = f'python3 {script} {field} {conf_file}'
        ut.printlog(cmd, log_field_file,
                    color = "blue", on_color = "on_white")
        os.system(cmd)

        print("\n\n")

    if 'calibration_stloc' in conf['run_steps']:

        script = os.path.join(pipeline_path, 'steps', 'calibration_stloc.py')
        cmd = f'python3 {script} {field} {conf_file}'
        ut.printlog(cmd, log_field_file,
                    color = "blue", on_color = "on_white")
        os.system(cmd)

        print("\n\n")

    if 'calibration_internal' in conf['run_steps']:

        script = os.path.join(pipeline_path, 'steps', 'calibration_internal.py')
        cmd = f'python3 {script} {field} {conf_file}'
        ut.printlog(cmd, log_field_file,
                    color = "blue", on_color = "on_white")
        os.system(cmd)

        print("\n\n")

    if 'calibration_gaiascale' in conf['run_steps']:

        script = os.path.join(pipeline_path, 'steps', 'calibration_gaiascale.py')
        cmd = f'python3 {script} {field} {conf_file}'
        ut.printlog(cmd, log_field_file,
                    color = "blue", on_color = "on_white")
        os.system(cmd)

        print("\n\n")

    if 'calibration_diagnostic' in conf['run_steps']:

        script = os.path.join(pipeline_path, 'steps', 'calibration_diagnostic.py')
        cmd = f'python3 {script} {field} {conf_file}'
        ut.printlog(cmd, log_field_file,
                    color = "blue", on_color = "on_white")
        os.system(cmd)

        print("\n\n")

    # It only needs to check for calibration_external to know if any calibration
    # is being run.
    if 'calibration_external' in conf['run_steps']:

        script = os.path.join(pipeline_path, 'steps', 'calibration_finalzp.py')
        cmd = f'python3 {script} {field} {conf_file}'
        ut.printlog(cmd, log_field_file,
                    color = "blue", on_color = "on_white")
        os.system(cmd)

        print("\n\n")

    if 'calibration_catalog' in conf['run_steps']:

        script = os.path.join(pipeline_path, 'steps', 'calibration_catalog.py')
        cmd = f'python3 {script} {field} {conf_file}'
        ut.printlog(cmd, log_field_file,
                    color = "blue", on_color = "on_white")
        os.system(cmd)

        print("\n\n")
    
    if 'quality_check' in conf['run_steps']:

        if "--dual_quality_check" in sys.argv:
            script = os.path.join(pipeline_path, 'quality_check.py')
        else:
            if conf['calibration_photometry'].lower() in ['single', 'dual']:
                script = os.path.join(pipeline_path, 'quality_check.py')
            elif conf['calibration_photometry'].lower() in ['psf']:
                script = os.path.join(pipeline_path, 'quality_check_PSF.py')

        cmd = f'python3 {script} {field} {conf_file}'
        ut.printlog(cmd, log_field_file,
                    color = "blue", on_color = "on_white")
        os.system(cmd)

        print("\n\n")

    ############################################################################

    ###################
    # Remove fits files

    if run_photometry and conf['remove_fits'] and not ('--nofits' in sys.argv) \
                      and not ('--keepfits' in sys.argv):

        for filt in conf['filters']:
            image_name = f'{field}_{filt}_swp.fz'
            image_fz = os.path.join(field_images_path, image_name)
            image_fits = image_fz.replace(".fz", ".fits")

            # Only remove .fits if .fz exists:
            if os.path.exists(image_fz) and os.path.exists(image_fits):

                ut.printlog(f"Removing image {image_fits}", log_field_file,
                             color = "red")

                cmd = f"rm {image_fits}"
                ut.printlog(cmd, log_field_file, color = "yellow")
                os.system(cmd)

            if conf['use_weight']:

                weight_fz = image_fz.replace(".fz", "weight.fz")
                weight_fits = weight_fz.replace(".fz", ".fits")

                if os.path.exists(weight_fz) and os.path.exists(weight_fits):
                    ut.printlog(f"Removing weight {weight_fits}",
                                log_field_file, color = "yellow")

                    cmd = f"rm {weight_fits}"
                    ut.printlog(cmd, log_field_file, color = "red")
                    os.system(cmd)

    #################################
    # Remove all images, including fz

    if os.path.exists(field_images_path) and conf["remove_images"]:
        ut.printlog(f"Removing images, including .fz",
                    log_field_file, color = "yellow")

        cmd = f"rm -r {field_images_path}"
        ut.printlog(cmd, log_field_file, color = "red", attrs = ["bold"])
        os.system(cmd)

    #########################
    # Sync run_path and save_path field to save_path
    if conf['save_path'] != conf['run_path']:
        
        cmd = f"rsync --update -haz --progress {field_path} {conf['save_path']}"
        ut.printlog(cmd, log_field_file, color = "blue")
        os.system(cmd)
        
        cmd = f"rm -r {field_path}"
        ut.printlog(cmd, log_field_file, color = "red", attrs=["bold"])
        os.system(cmd)
    



# -*- coding: utf-8 -*-

# ******************************************************************************
#                          S-PLUS CALIBRATION PIPELINE
#                             photometry_single.py
#                      aperture photometry in single mode
# ******************************************************************************

"""
Runs SExtractor in single mode for one S-PLUS field and a list of filters.

The S-PLUS field is given as the first command line argument.

The set of filters, location of S-PLUS images, default SExtractor configuration
and param files, and the output path must be set in the configuration file
given as the second command line argument

--------------------------------------------------------------------------------
   FUNCTIONS:
--------------------------------------------------------------------------------

create_photometry_path()
generate_sexconfig_files()
run_sex_single()
single_diagnostic_plots()

--------------------------------------------------------------------------------
   COMMENTS:
--------------------------------------------------------------------------------
Ideally this script should only be run through the pipeline.py script.

--------------------------------------------------------------------------------
   USAGE:
--------------------------------------------------------------------------------
$python3 photometry_single.py *field_name* *config_file*

----------------
"""

################################################################################
# Import external packages

import os
import sys

steps_path = os.path.split(__file__)[0]
pipeline_path = os.path.split(steps_path)[0]
spluscalib_path = os.path.split(pipeline_path)[0]

sys.path.append(spluscalib_path)

################################################################################
# Import spluscalib packages

from spluscalib import utils as ut

################################################################################
# Read parameters

field     = sys.argv[1]
conf_file = sys.argv[2]

conf = ut.pipeline_conf(conf_file)

################################################################################
# Get directories

field_path      = os.path.join(conf['run_path'], field)
photometry_path = os.path.join(field_path, 'Photometry')
single_path     = os.path.join(photometry_path, 'single')
sexconf_path    = os.path.join(single_path, 'sexconf')
catalogs_path   = os.path.join(single_path, 'catalogs')

images_path     = os.path.join(field_path, 'Images')

log_path        = os.path.join(photometry_path, 'logs')

################################################################################
# Create Photometry directory

ut.makedir(field_path)
ut.makedir(photometry_path)
ut.makedir(log_path)

################################################################################
# Initiate log file

log_file_name = os.path.join(log_path, 'photometry_single.log')
log_file_name = ut.gen_logfile_name(log_file_name)
log_file = os.path.join(photometry_path, log_file_name)

with open(log_file, "w") as log:
    log.write("")

################################################################################
# Log configuration

ut.printlog("Single mode photometry parameters:", log_file)

single_mode_params = ['run_path', 'filters', 'path_to_images', 'inst_zp',
                      'use_weight', 'path_to_sex', 'sex_config', 'sex_param',
                      'sex_XY_correction', 'XY_correction_maps_path',
                      'XY_correction_xbins', 'XY_correction_ybins',
                      ]

for param in single_mode_params:
    try:
        ut.printlog(f"{param}: {conf[param]}", log_file)
    except KeyError:
        ut.printlog(f"{param}: NONE", log_file)

################################################################################
# Begin script

# ***************************************************
#    If photometry path does not exist -> create
# ***************************************************


def create_photometry_path():

    print("")
    ut.printlog(('********** '
                 'Generating Single Photometry paths '
                 '**********'),
                 log_file)
    print("")

    # Create single mode photometry path
    ut.makedir(single_path)

    # Create single mode sexconf path
    ut.makedir(sexconf_path)

    # Create single mode catalogs path
    ut.makedir(catalogs_path)


create_photometry_path()


# ***************************************************
#    Generate SExconfig files
# ***************************************************

def generate_sexconfig_files():

    """
    Generate single mode sexconfig file for all filters
    """

    print("")
    ut.printlog(('********** '
                 'Generating Single Photometry config files '
                 '**********'),
                 log_file)
    print("")

    # Default SExconfig and SExparam files
    default_sexconfig = conf['sex_config']
    default_sexparam  = conf['sex_param']

    # Generate configuration for each filter
    for filt in conf['filters']:

        # Generate location of config file
        conf_file_name = f"{field}_{filt}_single.sex"
        save_file = os.path.join(sexconf_path, conf_file_name)

        # Generate location of output catalog file
        catalog_name = f"sex_{field}_{filt}_single.fits"
        catalog_file = os.path.join(catalogs_path, catalog_name)

        # Get location of image file
        image_name = f"{field}_{filt}_swp.fits"
        image_file = os.path.join(images_path, image_name)

        if not os.path.exists(save_file):

            ut.get_sex_config(save_file = save_file,
                              default_sexconfig = default_sexconfig,
                              default_sexparam  = default_sexparam,
                              catalog_file      = catalog_file,
                              image_file        = image_file,
                              inst_zp           = conf['inst_zp'],
                              path_to_sex       = conf['path_to_sex'],
                              use_weight        = conf['use_weight'],
                              mode              = 'single')

            ut.printlog(("Generated single mode sexconfig for field "
                         f"{field}, filter {filt}"), log_file)

        else:
            ut.printlog(("Single mode sexconfig already obtained for field "
                         f"{field}, filter {filt}"), log_file)

    # Copy sexparam file to field config directory (for proper documentation)
    cmd = f'cp {default_sexparam} {sexconf_path}'
    ut.printlog("$ "+cmd, log_file)
    os.system(cmd)


generate_sexconfig_files()

# ***************************************************
#    Run SExtractor in Single Mode
# ***************************************************


def run_sex_single():

    """
    Run SExtractor in Single mode for all filters
    """

    print("")
    ut.printlog(('********** '
                 'Running SeX on single mode '
                 '**********'),
                 log_file)
    print("")

    for filt in conf['filters']:

        catalog_name = f"sex_{field}_{filt}_single.fits"
        catalog_file = os.path.join(catalogs_path, catalog_name)

        if not os.path.exists(catalog_file):

            # Get location of image file
            image_name = f"{field}_{filt}_swp.fits"
            image_file = os.path.join(images_path, image_name)

            # Get location of sexconfig file
            conf_file_name = f"{field}_{filt}_single.sex"
            sexconf_file = os.path.join(sexconf_path, conf_file_name)

            # Run SExtractor
            sex_location = os.path.join(conf['path_to_sex'], 'src', 'sex')

            cmd = f"{sex_location} {image_file} -c {sexconf_file}"
            ut.printlog("$ "+cmd, log_file)
            os.system(cmd)

        else:
            ut.printlog(("Single mode photometry already obtained for field "
                         f"{field}, filter {filt}"), log_file)


run_sex_single()


# ***************************************************
#    Single diagnostic plots
# ***************************************************

def single_diagnostic_plots():

    """
    Makes diagnostic plots of the resulting catalogs
    """

    print("")
    ut.printlog(('********** '
                 'Making diagnostic plots for single mode photometry'
                 '**********'),
                log_file)
    print("")

    for filt in conf['filters']:

        # Generate location of output catalog file
        catalog_name = f"sex_{field}_{filt}_single.fits"
        catalog_file = os.path.join(catalogs_path, catalog_name)

        save_name = f"{field}_{filt}_single.png"
        save_file = os.path.join(catalogs_path, save_name)

        conf_file_name = f"{field}_{filt}_single.sex"
        sexconf_file = os.path.join(sexconf_path, conf_file_name)

        if not os.path.exists(save_file):

            ut.plot_sex_diagnostic(catalog   = catalog_file,
                                   save_file = save_file,
                                   s2ncut    = conf['apercorr_s2ncut'],
                                   starcut   = conf['apercorr_starcut'],
                                   sexconf   = sexconf_file,
                                   filt      = filt)

            ut.printlog(f"Created file {save_file}", log_file)

        else:
            ut.printlog(("Diagnostic plots already made for field "
                         f"{field}, filter {filt}"), log_file)


single_diagnostic_plots()

ut.printlog("COMPLETED SINGLE MODE PHOTOMETRY", log_file)



# -*- coding: utf-8 -*-

# ******************************************************************************
#                          S-PLUS CALIBRATION PIPELINE
#                               photometry_id.py
#                  Assign detection IDs to instrumental catalogs
# ******************************************************************************

"""
Assign detection IDs to instrumental catalogs. In dual mode aperture photometry,
IDs in the filter catalogs matches the IDs of the detection image. In PSF and
single mode aperture photometry there is no correlation between IDs of different
filters

The S-PLUS field is given as the first command line argument.

Configuration parameters are set in the configuration file given as the second
command line argument

--------------------------------------------------------------------------------
   FUNCTIONS:
--------------------------------------------------------------------------------
create_id_path()
include_filter_id_in_single_mode_catalogs()
extract_filter_id_radec_from_single_catalogs()
crossmatch_filter_id_catalogs_single()
create_field_id_catalog_single()
include_field_id_in_single_mode_catalogs()
remove_temporary_single_mode_filter_id_files()
generate_dual_mode_field_ids()
include_field_id_in_dual_mode_catalogs()
include_field_id_in_detection_catalog()
extract_filter_id_radec_from_psf_catalogs()
crossmatch_filter_id_catalogs_psf()
create_field_id_catalog_psf()
include_field_id_in_psf_mode_catalogs()

--------------------------------------------------------------------------------
   COMMENTS:
--------------------------------------------------------------------------------
Ideally this script should only be run through the pipeline.py script.

Assumes that photometry_psf.py and/or photometry_single.py and/or
photometry_dual.py (and optionally correction_xy.py) and correction_aper has
already been run for this field.

--------------------------------------------------------------------------------
   USAGE:
--------------------------------------------------------------------------------
$python3 photometry_id.py *field_name* *config_file*

----------------
"""

################################################################################
# Import external packages

import os
import sys

steps_path = os.path.split(__file__)[0]
pipeline_path = os.path.split(steps_path)[0]
spluscalib_path = os.path.split(pipeline_path)[0]

sys.path.append(spluscalib_path)

################################################################################
# Import spluscalib packages

from spluscalib import utils as ut

################################################################################
# Read parameters

field     = sys.argv[1]
conf_file = sys.argv[2]

conf = ut.pipeline_conf(conf_file)


################################################################################
# Get directories

field_path      = os.path.join(conf['run_path'], field)
photometry_path = os.path.join(field_path, 'Photometry')

single_path         = os.path.join(photometry_path, 'single')
single_catalog_path = os.path.join(single_path, 'catalogs')
single_xycorr_path = os.path.join(single_path, 'xy_correction')
single_apercorr_path = os.path.join(single_path, 'aper_correction')
single_id_path  = os.path.join(single_path, 'id')

dual_path         = os.path.join(photometry_path, 'dual')
dual_catalog_path = os.path.join(dual_path, 'catalogs')
dual_xycorr_path = os.path.join(dual_path, 'xy_correction')
dual_apercorr_path = os.path.join(dual_path, 'aper_correction')
dual_detection_path = os.path.join(dual_path, 'detection')
dual_id_path  = os.path.join(dual_path, 'id')

psf_path         = os.path.join(photometry_path, 'psf')
psf_catalog_path = os.path.join(psf_path, 'catalogs')
psf_xycorr_path = os.path.join(psf_path, 'xy_correction')
psf_id_path  = os.path.join(psf_path, 'id')

images_path     = os.path.join(field_path, 'Images')

log_path        = os.path.join(photometry_path, 'logs')


################################################################################
# Initiate log file

ut.makedir(log_path)

log_file_name = os.path.join(log_path, 'photometry_id.log')
log_file_name = ut.gen_logfile_name(log_file_name)
log_file = os.path.join(photometry_path, log_file_name)

with open(log_file, "w") as log:
    log.write("")

################################################################################
# Log configuration

ut.printlog("Single mode photometry parameters:", log_file)

master_phot_params = ['run_path', 'filters', 'data_release_name']

for param in master_phot_params:
    try:
        ut.printlog(f"{param}: {conf[param]}", log_file)
    except KeyError:
        ut.printlog(f"{param}: NONE", log_file)

################################################################################
# Begin script

# ***************************************************
#    If photometry path does not exist -> create
# ***************************************************


def create_id_path():

    print("")
    ut.printlog(('********** '
                 'Generating id paths '
                 '**********'),
                 log_file)
    print("")

    # Create single mode id path
    if 'photometry_single' in conf['run_steps']:
        ut.makedir(single_id_path)

    # Create dual mode id path
    if 'photometry_dual' in conf['run_steps']:
        ut.makedir(dual_id_path)

    # Create psf mode id path
    if 'photometry_psf' in conf['run_steps']:
        ut.makedir(psf_id_path)


create_id_path()


# ***************************************************
#    Assign single mode filter IDs
# ***************************************************

def include_filter_id_in_single_mode_catalogs():
    print("")
    ut.printlog(('********** '
                 'Including filter IDs in single mode catalogs '
                 '**********'),
                log_file)
    print("")

    for filt in conf['filters']:

        if "correction_aper" in conf["run_steps"]:
            catalog_name = f"sex_{field}_{filt}_single_apercorr.fits"
            catalog = os.path.join(single_apercorr_path, catalog_name)
        else:
            if "correction_xy" in conf["run_steps"]:
                catalog_name = f"sex_{field}_{filt}_single_xycorr.fits"
                catalog = os.path.join(single_xycorr_path, catalog_name)
            else:
                catalog_name = f"sex_{field}_{filt}_single.fits"
                catalog = os.path.join(single_catalog_path, catalog_name)

        save_name = f'sex_{field}_{filt}_single_tmp.fits'
        save_file = os.path.join(single_id_path, save_name)

        future_name = f'{field}_{filt}_id+radec_single.fits'
        future_file = os.path.join(single_id_path, future_name)

        if not os.path.exists(save_file) and not os.path.exists(future_file):
            ut.assign_single_mode_filter_id(catalog   = catalog,
                                            filt      = filt,
                                            save_file = save_file,
                                            field     = field,
                                     drname    = conf["data_release_name"])

            ut.printlog((f"Assigned Filter IDs for field {field}, "
                         f"filter {filt} (single mode)"), log_file)

        else:
            ut.printlog((f"Filter IDs already assigned for field {field}, "
                         f"filter {filt} (single mode)"), log_file)


if 'photometry_single' in conf['run_steps']:
    include_filter_id_in_single_mode_catalogs()

# ***************************************************
#    Extract filt_ID+RADEC catalogs
# ***************************************************

def extract_filter_id_radec_from_single_catalogs():
    print("")
    ut.printlog(('********** '
                 'Extracting filter IDs, RA, DEC from single mode catalogs '
                 '**********'),
                log_file)
    print("")

    for filt in conf['filters']:

        catalog_name = f'sex_{field}_{filt}_single_tmp.fits'
        catalog = os.path.join(single_id_path, catalog_name)

        save_name = f'{field}_{filt}_id+radec_single.fits'
        save_file = os.path.join(single_id_path, save_name)

        if not os.path.exists(save_file):
            ut.extract_filt_id_single(catalog   = catalog,
                                      filt      = filt,
                                      save_file = save_file)

            ut.printlog((f"Extracted filter IDs for field {field}, "
                         f"filter {filt} (single mode)"), log_file)

        else:
            ut.printlog((f"Filter IDs already extracted for field {field}, "
                         f"filter {filt} (single mode)"), log_file)


if 'photometry_single' in conf['run_steps']:
    extract_filter_id_radec_from_single_catalogs()


# ***************************************************
#    Crossmatch filt catalogs single
# ***************************************************

def crossmatch_filter_id_catalogs_single():
    print("")
    ut.printlog(('********** '
                 'Crossmatching filter ID catalogs (single mode) '
                 '**********'),
                log_file)
    print("")

    save_name = f'{field}_all_id+radec_single.fits'
    save_file = os.path.join(single_id_path, save_name)

    if not os.path.exists(save_file):

        cmd  = f"java -Xss4m -jar {conf['path_to_stilts']} tmatchn "
        cmd += f"nin={len(conf['filters'])} matcher=sky params=3 "
        cmd += f"multimode=group "

        j = 1
        for filt in conf['filters']:
            catalog_name = f'{field}_{filt}_id+radec_single.fits'
            catalog = os.path.join(single_id_path, catalog_name)

            filt_standard = ut.translate_filter_standard(filt)
            valuesN = f"'RA_{filt_standard} DEC_{filt_standard}'"

            cmd += f"in{j}={catalog} ifmt{j}=fits values{j}={valuesN} "
            cmd += f"join{j}=always "
            j += 1

        cmd += f"out={save_file} ofmt=fits"

        ut.printlog(cmd, log_file)
        os.system(cmd)

        ut.printlog((f"Crossmatched Filter IDs for field {field}, "
                     f"(single mode)"), log_file)

    else:
        ut.printlog((f"Filter IDs already crossmatched for field {field}, "
                     f"(single mode)"), log_file)


if 'photometry_single' in conf['run_steps']:
    crossmatch_filter_id_catalogs_single()


# ***************************************************
#    Create field ID catalog for single mode
# ***************************************************

def create_phot_id_catalog_single():

    print("")
    ut.printlog(('********** '
                 'Creating PHOT ID catalog (single mode) '
                 '**********'),
                log_file)
    print("")

    catalog_name = f'{field}_all_id+radec_single.fits'
    catalog_file = os.path.join(single_id_path, catalog_name)

    save_name = f'{field}_PHOT_ID_single.fits'
    save_file = os.path.join(single_id_path, save_name)

    if not os.path.exists(save_file):
        ut.generate_phot_id(catalog=catalog_file,
                             filters=conf["filters"],
                             field=field,
                             drname=conf["data_release_name"],
                             mode="single",
                             save_file = save_file)

        ut.printlog((f"Generated PHOT IDs for field {field}, "
                     f"(single mode)"), log_file)

    else:
        ut.printlog((f"PHOT IDs already generated for field {field}, "
                     f"(single mode)"), log_file)


if 'photometry_single' in conf['run_steps']:
    create_phot_id_catalog_single()


# ***************************************************
#    Generate dual mode PHOT IDs catalog
# ***************************************************

def create_phot_id_catalog_dual():
    print("")
    ut.printlog(('********** '
                 'Creating PHOT ID catalog (dual mode) '
                 '**********'),
                log_file)
    print("")

    catalog_name = f"sex_{field}_detection_catalog.fits"
    catalog = os.path.join(dual_detection_path, catalog_name)

    save_name = f'{field}_PHOT_ID_dual.fits'
    save_file = os.path.join(dual_id_path, save_name)

    if not os.path.exists(save_file):
        ut.generate_dual_mode_phot_id(catalog=catalog,
                                  save_file=save_file,
                                  field=field,
                                  drname=conf["data_release_name"])

        ut.printlog(f"Generated dual PHOT IDs for field {field}", log_file)

    else:
        ut.printlog(f"Dual PHOT IDs already exist for field {field}", log_file)


if 'photometry_dual' in conf['run_steps']:
    create_phot_id_catalog_dual()


# ***************************************************
#    Extract filt_ID+RADEC catalogs (PSF)
# ***************************************************

def extract_filter_id_radec_from_psf_catalogs():
    print("")
    ut.printlog(('********** '
                 'Extracting filter IDs, RA, DEC from psf catalogs '
                 '**********'),
                log_file)
    print("")

    for filt in conf['filters']:

        if "correction_xy" in conf["run_steps"]:
            catalog_name = f'{field}_{filt}_psf_xycorr.cat'
            catalog_file = os.path.join(psf_xycorr_path, catalog_name)
        else:
            catalog_name = f'{field}_{filt}_psf.cat'
            catalog_file = os.path.join(psf_catalog_path, catalog_name)

        save_name = f'{field}_{filt}_id+radec_psf.fits'
        save_file = os.path.join(psf_id_path, save_name)

        if not os.path.exists(save_file):
            ut.extract_filt_id_psf(catalog   = catalog_file,
                                   filt      = filt,
                                   save_file = save_file)

            ut.printlog((f"Extracted filter IDs for field {field}, "
                         f"filter {filt} (psf mode)"), log_file)

        else:
            ut.printlog((f"Filter IDs already extracted for field {field}, "
                         f"filter {filt} (psf mode)"), log_file)


if 'photometry_psf' in conf['run_steps']:
    extract_filter_id_radec_from_psf_catalogs()


# ***************************************************
#    Crossmatch filt catalogs (psf)
# ***************************************************

def crossmatch_filter_id_catalogs_psf():
    print("")
    ut.printlog(('********** '
                 'Crossmatching filter ID catalogs (psf mode) '
                 '**********'),
                log_file)
    print("")

    save_name = f'{field}_all_id+radec_psf.fits'
    save_file = os.path.join(psf_id_path, save_name)

    if not os.path.exists(save_file):

        cmd  = f"java -Xss8m -jar {conf['path_to_stilts']} tmatchn "
        cmd += f"nin={len(conf['filters'])} matcher=sky params=3 "
        cmd += f"multimode=group "

        j = 1
        for filt in conf['filters']:
            catalog_name = f'{field}_{filt}_id+radec_psf.fits'
            catalog = os.path.join(psf_id_path, catalog_name)

            filt_standard = ut.translate_filter_standard(filt)
            valuesN = f"'RA_{filt_standard} DEC_{filt_standard}'"

            cmd += f"in{j}={catalog} ifmt{j}=fits values{j}={valuesN} "
            cmd += f"join{j}=always "
            j += 1

        cmd += f"out={save_file} ofmt=fits"

        ut.printlog(cmd, log_file)
        os.system(cmd)

        ut.printlog((f"Crossmatched Filter IDs for field {field}, "
                     f"(psf mode)"), log_file)

    else:
        ut.printlog((f"Filter IDs already crossmatched for field {field}, "
                     f"(psf mode)"), log_file)


if 'photometry_psf' in conf['run_steps']:
    crossmatch_filter_id_catalogs_psf()


# ***************************************************
#    Create PHOT ID catalog for psf mode
# ***************************************************

def create_phot_id_catalog_psf():

    print("")
    ut.printlog(('********** '
                 'Creating PHOT ID catalog (psf mode) '
                 '**********'),
                log_file)
    print("")

    catalog_name = f'{field}_all_id+radec_psf.fits'
    catalog_file = os.path.join(psf_id_path, catalog_name)

    save_name = f'{field}_PHOT_ID_psf.fits'
    save_file = os.path.join(psf_id_path, save_name)

    if not os.path.exists(save_file):
        ut.generate_phot_id(catalog=catalog_file,
                            filters=conf["filters"],
                            field=field,
                            drname=conf["data_release_name"],
                            mode="psf",
                            save_file = save_file)

        ut.printlog((f"Generated PHOT IDs for field {field}, "
                     f"(psf mode)"), log_file)

    else:
        ut.printlog((f"PHOT IDs already generated for field {field}, "
                     f"(psf mode)"), log_file)

if 'photometry_psf' in conf['run_steps']:
    create_phot_id_catalog_psf()


# ***************************************************
#    Combining PHOT ID catalogs all modes
# ***************************************************

def combining_phot_ID_catalogs():

    print("")
    ut.printlog(('********** '
                 'Combining PHOT ID catalogs '
                 '**********'),
                log_file)
    print("")

    single_name = f'{field}_PHOT_ID_single.fits'
    single_file = os.path.join(single_id_path, single_name)

    dual_name = f'{field}_PHOT_ID_dual.fits'
    dual_file = os.path.join(dual_id_path, dual_name)

    psf_name = f'{field}_PHOT_ID_psf.fits'
    psf_file = os.path.join(psf_id_path, psf_name)

    save_name = f'{field}_PHOT_ID_all.fits'
    save_file = os.path.join(photometry_path, save_name)

    if not os.path.exists(save_file):
        cmd  = f"java -Xss4m -jar {conf['path_to_stilts']} tmatchn "
        cmd += f"matcher=sky params=5 "
        cmd += f"multimode=group "

        j = 0
        last_mode = ""

        if "photometry_dual" in conf["run_steps"]:
            j += 1
            valuesN = "'PHOT_ID_RA_dual PHOT_ID_DEC_dual'"
            cmd += f"in{j}={dual_file} ifmt{j}=fits values{j}={valuesN} "
            cmd += f"join{j}=always "

            last_mode = "dual"

        if "photometry_single" in conf["run_steps"]:
            j += 1
            valuesN = "'PHOT_ID_RA_single PHOT_ID_DEC_single'"
            cmd += f"in{j}={single_file} ifmt{j}=fits values{j}={valuesN} "
            cmd += f"join{j}=always "

            last_mode = "single"

        if "photometry_psf" in conf["run_steps"]:
            j += 1
            valuesN = "'PHOT_ID_RA_psf PHOT_ID_DEC_psf'"
            cmd += f"in{j}={psf_file} ifmt{j}=fits values{j}={valuesN} "
            cmd += f"join{j}=always "

            last_mode = "psf"

        cmd += f"nin={j} out={save_file} ofmt=fits"

        if j > 1:
            ut.printlog(cmd, log_file)
            os.system(cmd)

        else:
            if last_mode == "dual":
                phot_id_catalog = dual_file
            elif last_mode == "single":
                phot_id_catalog = single_file
            elif last_mode == "psf":
                phot_id_catalog = psf_file
            else:
                raise ValueError(f"{last_mode} is not an acceptable value.")

            cmd = f"cp {phot_id_catalog} {save_file}"
            ut.printlog(cmd, log_file)
            os.system(cmd)

        ut.printlog(f"Combined PHOT IDs for field {field}, ", log_file)

    else:
        ut.printlog(f"PHOT IDs already combined for field {field}, ",
                    log_file)


combining_phot_ID_catalogs()

# ***************************************************
#    Generate FIELD_IDs for combined phot_ID catalogs
# ***************************************************

def creating_field_id_catalog():
    print("")
    ut.printlog(('********** '
                 'Creating Field ID catalog '
                 '**********'),
                log_file)
    print("")

    phot_id_catalog_name = f'{field}_PHOT_ID_all.fits'
    phot_id_catalog = os.path.join(photometry_path, phot_id_catalog_name)

    save_name = f'{field}_FIELD_ID.fits'
    save_file = os.path.join(photometry_path, save_name)

    if not os.path.exists(save_file):

        modes = []
        if "photometry_dual" in conf['run_steps']:
            modes.append("dual")
        if "photometry_single" in conf['run_steps']:
            modes.append("single")
        if "photometry_psf" in conf['run_steps']:
            modes.append("psf")

        ut.generate_field_id(catalog=phot_id_catalog,
                             field=field,
                             drname=conf["data_release_name"],
                             save_file=save_file,
                             modes = modes)

    else:
        ut.printlog(f"FIELD IDs already assigned for field {field}, ",
                    log_file)


creating_field_id_catalog()


# ***************************************************
#    Including MODE/FIELD IDs psf mode
# ***************************************************

def include_field_phot_ids_in_psf_catalogs():
    print("")
    ut.printlog(('********** '
                 'Including FIELD/PHOT ID in psf mode catalogs '
                 '**********'),
                log_file)
    print("")

    for filt in conf['filters']:

        if "correction_xy" in conf["run_steps"]:
            catalog_name = f'{field}_{filt}_psf_xycorr.cat'
            catalog_file = os.path.join(psf_xycorr_path, catalog_name)
        else:
            catalog_name = f'{field}_{filt}_psf.cat'
            catalog_file = os.path.join(psf_catalog_path, catalog_name)

        id_catalog_name = f'{field}_FIELD_ID.fits'
        id_catalog = os.path.join(photometry_path, id_catalog_name)

        save_name = f'{field}_{filt}_psf_withIDs.csv'
        save_file = os.path.join(psf_id_path, save_name)

        filt_standard = ut.translate_filter_standard(filt)
        values2 = f"'RA_{filt_standard} DEC_{filt_standard}'"

        if not os.path.exists(save_file):

            cmd = f"java -Xss4m -jar {conf['path_to_stilts']} tmatch2 "
            cmd += f"matcher=sky params=3 join=1and2 "
            cmd += f"in1={id_catalog} ifmt1=fits "
            cmd += f"values1='PHOT_ID_RA_psf PHOT_ID_DEC_psf' "
            cmd += f"in2={catalog_file} ifmt2=ascii "
            cmd += f"values2={values2} "
            cmd += f"out={save_file} ofmt=csv"

            print(cmd)
            os.system(cmd)

            ut.printlog((f"Assigned IDs for field {field}, "
                         f"filter {filt} (psf mode)"), log_file)

        else:
            ut.printlog((f"IDs already assigned for field {field}, "
                         f"filter {filt} (psf mode)"), log_file)


if 'photometry_psf' in conf['run_steps']:
    include_field_phot_ids_in_psf_catalogs()


# ***************************************************
#    Assign dual mode IDs
# ***************************************************

def include_field_phot_ids_in_dual_catalogs():
    print("")
    ut.printlog(('********** '
                 'Including FIELD/PHOT ID in dual mode catalogs '
                 '**********'),
                log_file)
    print("")

    for filt in conf['filters']:

        if "correction_aper" in conf["run_steps"]:
            catalog_name = f"sex_{field}_{filt}_dual_apercorr.fits"
            catalog = os.path.join(dual_apercorr_path, catalog_name)
        else:
            if "correction_xy" in conf["run_steps"]:
                catalog_name = f"sex_{field}_{filt}_dual_xycorr.fits"
                catalog = os.path.join(dual_xycorr_path, catalog_name)
            else:
                catalog_name = f"sex_{field}_{filt}_dual.fits"
                catalog = os.path.join(dual_catalog_path, catalog_name)

        id_catalog_name = f'{field}_FIELD_ID.fits'
        id_catalog = os.path.join(photometry_path, id_catalog_name)

        save_name = f'sex_{field}_{filt}_dual_withIDs.fits'
        save_file = os.path.join(dual_id_path, save_name)

        if not os.path.exists(save_file):
            cmd  = f"java -Xss4m -jar {conf['path_to_stilts']} tmatch2 "
            cmd += f"matcher=sky params=3 join=1and2 "
            cmd += f"in1={id_catalog} ifmt1=fits "
            cmd += f"values1='PHOT_ID_RA_dual PHOT_ID_DEC_dual' "
            cmd += f"in2={catalog} ifmt2=fits "
            cmd += f"values2='ALPHA_J2000 DELTA_J2000' "
            cmd += f"out={save_file} ofmt=fits"

            print(cmd)
            os.system(cmd)

            ut.printlog((f"Assigned IDs for field {field}, "
                         f"filter {filt} (dual mode)"), log_file)

        else:
            ut.printlog((f"IDs already assigned for field {field}, "
                         f"filter {filt} (dual mode)"), log_file)


if 'photometry_dual' in conf['run_steps']:
    include_field_phot_ids_in_dual_catalogs()


# ***************************************************
#    Include FIELD_ID in detection catalog
# ***************************************************

def include_field_phot_ids_in_detection_catalog():
    print("")
    ut.printlog(('********** '
                 'Including Field IDs in detection catalog '
                 '**********'),
                log_file)
    print("")

    catalog_name = f"sex_{field}_detection_catalog.fits"
    catalog = os.path.join(dual_detection_path, catalog_name)

    id_catalog_name = f'{field}_FIELD_ID.fits'
    id_catalog = os.path.join(photometry_path, id_catalog_name)

    save_name = f'sex_{field}_detection_withIDs.fits'
    save_file = os.path.join(dual_id_path, save_name)

    if not os.path.exists(save_file):
        cmd = f"java -Xss4m -jar {conf['path_to_stilts']} tmatch2 "
        cmd += f"matcher=sky params=3 join=1and2 "
        cmd += f"in1={id_catalog} ifmt1=fits "
        cmd += f"values1='PHOT_ID_RA_dual PHOT_ID_DEC_dual' "
        cmd += f"in2={catalog} ifmt2=fits "
        cmd += f"values2='ALPHA_J2000 DELTA_J2000' "
        cmd += f"out={save_file} ofmt=fits"

        os.system(cmd)

        ut.printlog(f"Assigned IDs for field {field}, detection catalog ", log_file)

    else:
        ut.printlog((f"IDs already assigned for field {field}, "
                     f"detection catalog"), log_file)


if 'photometry_dual' in conf['run_steps']:
    include_field_phot_ids_in_detection_catalog()


# ***************************************************
#    Assign single mode IDs
# ***************************************************

def include_field_phot_ids_in_single_catalogs():
    print("")
    ut.printlog(('********** '
                 'Including FIELD/PHOT IDs in single mode catalogs '
                 '**********'),
                log_file)
    print("")

    for filt in conf['filters']:

        catalog_name = f'sex_{field}_{filt}_single_tmp.fits'
        catalog = os.path.join(single_id_path, catalog_name)

        id_catalog_name = f'{field}_FIELD_ID.fits'
        id_catalog = os.path.join(photometry_path, id_catalog_name)

        save_name = f'sex_{field}_{filt}_single_withIDs.fits'
        save_file = os.path.join(single_id_path, save_name)

        if not os.path.exists(save_file):
            cmd  = f"java -Xss4m -jar {conf['path_to_stilts']} tmatch2 "
            cmd += f"matcher=sky params=3 join=1and2 "
            cmd += f"in1={id_catalog} ifmt1=fits "
            cmd += f"values1='PHOT_ID_RA_single PHOT_ID_DEC_single' "
            cmd += f"in2={catalog} ifmt2=fits "
            cmd += f"values2='ALPHA_J2000 DELTA_J2000' "
            cmd += f"out={save_file} ofmt=fits"

            print(cmd)
            os.system(cmd)

            ut.printlog((f"Assigned IDs for field {field}, "
                         f"filter {filt} (single mode)"), log_file)

        else:
            ut.printlog((f"IDs already assigned for field {field}, "
                         f"filter {filt} (single mode)"), log_file)


if 'photometry_dual' in conf['run_steps']:
    include_field_phot_ids_in_single_catalogs()


# ***************************************************
#    Removing temporary files
# ***************************************************

def remove_temporary_single_mode_filter_id_files():
    print("")
    ut.printlog(('********** '
                 'Removing temporary single mode Filter ID files '
                 '**********'),
                log_file)
    print("")

    for filt in conf['filters']:

        tmp_name = f'sex_{field}_{filt}_single_tmp.fits'
        tmp_file = os.path.join(single_id_path, tmp_name)

        catalog_name = f'sex_{field}_{filt}_single_withIDs.fits'
        catalog_file = os.path.join(single_id_path, catalog_name)

        if os.path.exists(catalog_file):
            if os.path.exists(tmp_file):
                cmd  = f"rm {tmp_file}"
                print(cmd)
                os.system(cmd)

                ut.printlog(f"Removed file {tmp_file}", log_file)


if 'photometry_single' in conf['run_steps']:
    remove_temporary_single_mode_filter_id_files()
